#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
"""
Create RNA biotype analysis using STAR and featureCounts
"""
## import useful modules
import os
import concurrent.futures
from termcolor import colored

## import my modules
from XICRA.config import set_config
from XICRA.scripts import STAR_caller, multiQC_report

from HCGB import sampleParser
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.files_functions as HCGB_files

## set to use as a module
## allow multiple software to map
## e.g. bowtie

#########################################
def run_mapping(options):
    print("TODO: Implement this module")

#########################################
def mapReads_module_STAR(options, pd_samples_retrieved, outdir_dict, Debug, 
                    max_workers_int, threads_job, start_time_partial, outdir, multimapping):
    
    """Organizes the mapping of the samples, executed in parallel.

    First, checks if the files needed to do the mapping (and posterior
    classification of the reads) have been provided by the user:
    fasta sequence + annotation or STAR index directory. 

    Then, sends the mapping in parallel for each sample calling mapReads_caller().    
    
    Finally, generate the MultiQC report  of the mapping for each sample.
    
    :param options: input parameters introduced by the user. See XICRA biotype -h.
    :param pd_samples_retrieved: data frame with the information of the samples
    :param outdir_dict: dictionary with the names of the samples and their files
    :param Debug: show extra information of the process
    :param max_workers_int: number of workers for each thread
    :param threads_job: number of threads to do the computation
    :param start_time_partial: time of the beggining of the process
    :param outdir: directory to store the results

    :type Debug: boolean
    :type max_workers_int: int
    :type threads_job: int 
    :type start_time_partial: int
    :type outdir: string


    :returns: None
    """

    # Group dataframe by sample name
    sample_frame = pd_samples_retrieved.groupby(["new_name"])
    
    ## options
    STAR_exe = set_config.get_exe("STAR", Debug=Debug)
    cwd_folder = os.path.abspath("./")
    folder=HCGB_files.create_subfolder('STAR_files', cwd_folder)

    ## For many samples it will have to load genome index in memory every time.
    ## For a unique sample it will not matter. Take care genome might stay in memory.
    ## Use before loop option LoadAndExit and then:
        ## in loop
        ## Use option LoadAndKeep, set shared memory > 30 Gb
    ## when finished loop Remove memory        
    
    ## check reference
    if (options.fasta):
        print ("+ Genome fasta file provided")
        print ("+ Create genomeDir for later usage...")
        options.fasta = os.path.abspath(options.fasta)
        
        ## create genomeDir
        options.genomeDir = STAR_caller.create_genomeDir(folder, STAR_exe, options.threads, options.fasta, options.limitRAM)
        
    elif (options.genomeDir):
        print ("+ genomeDir provided.")
        options.genomeDir = os.path.abspath(options.genomeDir)
        
    ## remove previous reference genome from memory
    print ("+ Remove genome in memory from previous call... (if any)")
    #STAR_caller.remove_Genome(STAR_exe, options.genomeDir, folder, options.threads)
    
    ## load reference genome
    #STAR_caller.load_Genome(folder, STAR_exe, options.genomeDir, options.threads)

    ## functions.time_functions.timestamp
    start_time_partial = HCGB_time.timestamp(start_time_partial)
    
    print ("+ Mapping sequencing reads for each sample retrieved...")

    ## send for each sample
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor:
        commandsSent = { executor.submit(mapReads_caller_STAR, sorted(cluster["sample"].tolist()), 
                                         outdir_dict[name[0]], name[0], threads_job, STAR_exe, 
                                         options.genomeDir, options.limitRAM, 
                                         Debug, multimapping): name[0] for name, cluster in sample_frame }

        for cmd2 in concurrent.futures.as_completed(commandsSent):
            details = commandsSent[cmd2]
            try:
                data = cmd2.result()
            except Exception as exc:
                print ('***ERROR:')
                print (cmd2)
                print('%r generated an exception: %s' % (details, exc))

    print ("\n\n+ Mapping reads has finished...")
    
    ## functions.time_functions.timestamp
    start_time_partial = HCGB_time.timestamp(start_time_partial)

    ## remove reference genome from memory
    #STAR_caller.remove_Genome(STAR_exe, options.genomeDir, folder, options.threads)
    
    ## functions.time_functions.timestamp
    start_time_partial = HCGB_time.timestamp(start_time_partial)

    ## retrieve mapping files
    if options.detached:
        input_dir = outdir
    else:
        input_dir = os.path.abspath(options.input)
    
    results_SampleParser = sampleParser.files.get_files(options, input_dir, "map", ["Aligned.sortedByCoord.out.bam"], options.debug, bam=True)
    del results_SampleParser['dirname']
    del results_SampleParser['ext']
    del results_SampleParser['tag']
    #del results_SampleParser['new_name']

    results_SampleParser = results_SampleParser.set_index('name')
    mapping_results = results_SampleParser.to_dict()['sample']

    ## Create mapping report    
    if (options.skip_report):
        print ("+ No report generation...")
    else:
        print ("\n+ Generating a report using MultiQC module.")
        outdir_report = HCGB_files.create_subfolder("report", outdir)

        ## get subdirs generated and call multiQC report module
        givenList = []
        print ("+ Detail information for each sample could be identified in separate folders:")
        
        ## call multiQC report module
        givenList = [ v for v in outdir_dict.values() ]
        my_outdir_list = set(givenList)

        ## debug message
        if (Debug):
            print (colored("\n**DEBUG: my_outdir_list for multiqc report **", 'yellow'))
            print (my_outdir_list)
            print ("\n")
        
        map_report = HCGB_files.create_subfolder("STAR", outdir_report)
        multiQC_report.multiQC_module_call(my_outdir_list, "STAR", map_report,"-dd 2")
        print ('\n+ A summary HTML report of each sample is generated in folder: %s' %map_report)

    return(start_time_partial, mapping_results)

#################################
def mapReads_caller_STAR(files, folder, name, threads, STAR_exe, genomeDir, limitRAM_option, Debug, multimapping):
    """Mapping of a given sample with STAR

    First, checks if the trimmed unjoined files exist for the sample and also
    if the calculation has not been done previously. 

    Executes STAR and generate the BAM file of the sample with mapReads script. 
    
    :param files: unjoined trimmed files of the sample
    :param folder: sample folder to store the results
    :param name: sample name
    :param threads: number of threads to do the computation
    :param STAR_exe: check the STAR software is available
    :param genomeDir: path to the genome directory to do the mappig
    :param limitRAM_option: limit RAM bytes to be used in the computation
    :param Debug: show extra information of the process
    :param multimapping: Flag to say whether to use multimapping reads or not
    
    :type folder: string
    :type name: string
    :type threads: int 
    :type start_exe: boolean
    :type genomeDir: string
    :type limitRAM_option: int
    :type Debug: boolean
    :type multimapping: boolean

    :returns: None
    """

    ## check if previously joined and succeeded
    filename_stamp = folder + '/.success'
    if os.path.isfile(filename_stamp):
        stamp = HCGB_time.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'STAR'), 'yellow'))
    else:
        ##
        if Debug:
            print ("\n** DEBUG: mapReads_caller options **\n")
            print ("folder: " + folder) 
            print ("name: " + name)
            print ("threads: " + str(threads))
            print ("STAR_exe: " + STAR_exe) 
            print ("genomeDir: " + genomeDir) 
            print ("limitRAM_option: " + str(limitRAM_option))
            print ("files: ")
            print (files)
            
        # Call STAR
        code_returned = STAR_caller.mapReads("LoadAndKeep", files, folder, name, STAR_exe, genomeDir, limitRAM_option, threads, Debug, multimapping)
        
        if (code_returned):
            HCGB_time.print_time_stamp(filename_stamp)
        else:
            print ("+ Mapping sample %s failed..." %name)
    
    ## return results
    return()
    



#!/usr/bin/env python3
############################################################
## Jose F. Sanchez, Marta Lopez & Lauro Sumoy             ##
## Copyright (C) 2019-2021 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
"""
This module performs piRNA analysis with several possible tools: 

Show a description of piRNA molecules here: http://pirnabank.ibab.ac.in/about.html
"""
## import useful modules
import os
import time
from termcolor import colored

## import my modules
from HCGB import sampleParser

import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.main_functions as HCGB_main

from XICRA.modules import help_XICRA, map, database
from XICRA.scripts import pilfer_caller

##############################################
def run_piRNA(options):
    '''
    '''
    
    ## init time
    start_time_total = time.time()

    ##################################
    ### show help messages if desired    
    ##################################
    if (options.help_format):
        ## help_format option
        help_XICRA.help_fastq_format()
    
    elif (options.help_project):
        ## information for project
        help_XICRA.project_help()
        exit()
    
    elif (options.help_piRNA):
        ## information for join reads
        help_XICRA.help_piRNA()
        exit()
        
    ## debugging messages
    global Debug
    if (options.debug):
        Debug = True
    else:
        Debug = False
        
    ### set as default paired_end mode
    if (options.single_end):
        options.pair = False
    else:
        options.pair = True
    
    HCGB_aes.pipeline_header('XICRA')
    HCGB_aes.boxymcboxface("piRNA analysis")
    print ("--------- Starting Process ---------")
    HCGB_time.print_time()

    ## absolute path for in & out
    input_dir = os.path.abspath(options.input)
    outdir=""

    ## set mode: project/detached
    if (options.detached):
        outdir = os.path.abspath(options.output_folder)
        options.project = False
    else:
        options.project = True
        outdir = input_dir        
    
    ## user software selection
    print ("+ Software for piRNA analysis selected:")
    print (options.soft_name)
    
    ## get files
    print ('+ Getting files from input folder... ')
    print ('+ Check if previously mapping was generated...')
    
    ## retrieve mapping files
    results_SampleParser = sampleParser.files.get_files(options, input_dir, "map", ["bam"], options.debug, bam=True)
    pd_samples_retrieved = results_SampleParser
    del results_SampleParser['dirname']
    del results_SampleParser['ext']
    del results_SampleParser['tag']

    results_SampleParser = results_SampleParser.set_index('name')
    mapping_results = results_SampleParser.to_dict()['sample']
    
    ##
    if Debug:
        HCGB_aes.debug_message("results_SampleParser", "yellow")
        print(results_SampleParser)
        
        HCGB_aes.debug_message("mapping_results", "yellow")
        print(mapping_results)        

    ## Call mapping if mapping_results is empty
    if not mapping_results:
        
        if Debug:
            HCGB_aes.debug_message("********************")
            HCGB_aes.debug_message("mapping samples")

        ## get files: use joined reads if paired-end data
        if options.pair:
            options.pair = False ## set paired-end to false for further prepocessing
            if options.noTrim:
                print ('+ Mode: fastq.\n+ Extension: ')
                print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
                pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
            else:
                print ('+ Mode: join.\n+ Extension: ')
                print ("[_joined.fastq]\n")
                pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "join", ['_joined.fastq'], options.debug)
        else:
            if options.noTrim:
                print ('+ Mode: fastq.\n+ Extension: ')
                print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
                pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
            else:
                print ('+ Mode: join.\n+ Extension: ')
                print ("[_joined.fastq]\n")
                pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "trim", ['_trim'], options.debug)
        
        ## debug message
        if (Debug):
            print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
            print (pd_samples_retrieved)
    
        ## generate output folder, if necessary
        print ("\n+ Create output folder(s):")
        if not options.project:
            HCGB_files.create_folder(outdir)
    
        ## for samples
        mapping_outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "map", options.debug)
        
        ## debug message
        if (Debug):
            print (colored("**DEBUG: mapping_outdir_dict **", 'yellow'))
            print (mapping_outdir_dict)
    
        # time stamp
        start_time_partial = HCGB_time.timestamp(start_time_total)
    
        ## optimize threads
        name_list = set(pd_samples_retrieved["new_name"].tolist())
        threads_job = HCGB_main.optimize_threads(options.threads, len(name_list)) ## threads optimization
        max_workers_int = int(options.threads/threads_job)
        
        ## debug message
        if (Debug):
            print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
            print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
            print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))
            
        ##############################################
        ## map Reads
        ##############################################
        
        ## TODO: Fix options require options.fasta, provided it or options.genomeDir
        (start_time_partial, mapping_results) = map.mapReads_module_STAR(options, pd_samples_retrieved, mapping_outdir_dict, 
                        options.debug, max_workers_int, threads_job, start_time_partial, outdir)
    
        ## debug message
        if (Debug):
             print (colored("**DEBUG: mapping_results **", 'yellow'))
             print (mapping_results)
        
        # time stamp
        start_time_partial = HCGB_time.timestamp(start_time_partial)

    ############################################################
    ## Download piRNA information: piRBase
    ############################################################
    if not (options.database):
        install_path =  os.path.dirname(os.path.realpath(__file__))
        options.database = os.path.join(install_path, "db_files") 
    else:
        options.database = os.path.abspath(options.database)
    
    print ("+ Create folder to store results: ", options.database)
    HCGB_files.create_folder(options.database)
    
    ## call database module and return information updated
    db_info = database.piRNA_info(options.database, options.species, Debug)

    ##############################################################
    ## Start the analysis
    ##############################################################
    ## generate output folder, if necessary
    if not options.project:
        print ("\n+ Create output folder(s):")
        HCGB_files.create_folder(outdir)
    
    ## for samples
    outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "piRNA", options.debug)
    
    ## Mapping is done, process bam files
    print ("+ Create a piRNA analysis for each sample retrieved...")    
    print("+ Intersecting annotation file and mapping file...")
    
    ## optimize threads
    name_list = set(pd_samples_retrieved["new_name"].tolist())
    #threads_job = HCGB_main.optimize_threads(options.threads, len(name_list)) ## threads optimization
    threads_job = options.threads ## So far all CPUs provided for each sample at a time
    max_workers_int = 1

    ## debug message
    if (Debug):
        print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
        print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
        print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))
    # Group dataframe by sample name
    sample_frame = pd_samples_retrieved.groupby(["new_name"])
    
    ## for each sample do piRNA analysis
    for name, cluster in sample_frame:
        if Debug:
            print(name[0])
            print(cluster)
            
        ## send for each sample        
        piRNA_analysis(sorted(cluster["sample"].tolist())[0], outdir_dict[name[0]], 
                       name[0], threads_job, 
                       options.soft_name, 
                       options.species, options.database, Debug)

    ## outdir
    outdir_report = HCGB_files.create_subfolder("report", outdir)
    expression_folder = HCGB_files.create_subfolder("piRNA", outdir_report)
    
    ## TODO: control if any other software implemented
    ## if pilfer
    pilfer_folder = HCGB_files.create_subfolder("pilfer", expression_folder)
    
    ## get results
    results_pilfer = sampleParser.files.get_files(options, input_dir, "piRNA", ["pilfer_clusters.bed"], options.debug, bam=False)
    del results_pilfer['dirname']
    del results_pilfer['ext']
    del results_pilfer['tag']
    results_pilfer = results_pilfer.set_index('name')
    pilfer_results = results_pilfer.to_dict()['sample']

    ## merge clusters generated
    print ("+ Summarize piRNA analysis for all samples...")
    pilfer_caller.pilfer_merge_samples_call(pilfer_results, pilfer_folder, options.debug)
        
    print ("\n\n+ piRNA analysis is finished...")
    print ("+ Let's summarize all results...")
    
    exit()
    
    print ("\n*************** Finish *******************")
    start_time_partial = HCGB_time.timestamp(start_time_total)
    
    print ("\n+ Exiting piRNA module.")
    return()

#########################################
def piRNA_analysis(bam_file, folder, name, threads, soft_list, species, database, Debug):
    
    if Debug:
        HCGB_aes.debug_message("BAM_FILE: " + bam_file, "yellow")
        HCGB_aes.debug_message("folder: " + folder, "yellow")
        HCGB_aes.debug_message("name: " + name, "yellow")
        HCGB_aes.debug_message("threads: " + str(threads), "yellow")
        HCGB_aes.debug_message("soft_list: ", "yellow")
        print(soft_list)
        HCGB_aes.debug_message("species: " + species, "yellow")
    
    ##
    for soft in soft_list:
        if (soft == "pilfer"):
            ## create pilfer analysis
            pilfer_folder = HCGB_files.create_subfolder('pilfer', folder)
            code_success = pilfer_caller.pilfer_module_call(pilfer_folder, name, bam_file, database, threads, species, Debug)   
            
            if not code_success:
                print ('** Some error ocurred during pilfer analysis for sample %s...' %name)
                return ()
                


#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez                                      ##
## Copyright (C) 2019-2024 Lauro Sumoy Lab, IGTP, Spain ##
##########################################################
"""
Trims sequence adapters within fastq reads.
"""
## import useful modules
import os
import re
import time
import concurrent.futures
from termcolor import colored

## import my modules
from XICRA import __version__ as pipeline_version
from XICRA.scripts import multiQC_report
from XICRA.scripts import cutadapt_caller
from XICRA.modules import help_XICRA
from XICRA.modules import qc
from HCGB import sampleParser
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.info_functions as HCGB_info
import HCGB.functions.main_functions as HCGB_main

##############################################
def run_trim(options):
    """Main function of the module, organizes the trimming process.

    First, checks if the adapter(s) sequence(s) have been provided by the user:
    - adapters_a or adapters_A
    If there is no adapter sequence provided the process will be stopped.  

    If the adapters have been introduced, it calls cutadapt_caller() for each sample in parallel.
    Finally, generates a report using MultiQC module if desired.
    
    :param options: input parameters introduced by the user. See XICRA trim -h.

    :returns: None
    """
    ## init time
    start_time_total = time.time()

    ##################################
    ### show help messages if desired    
    ##################################
    if (options.help_format):
        ## help_format option
        help_XICRA.help_fastq_format()
    elif (options.help_trimm_adapters):
        ## help on trimm adapters
        help_XICRA.print_help_adapters()
        exit()
    elif (options.help_project):
        ## information for project
        help_XICRA.project_help()
        exit()
    elif (options.help_multiqc):
        ## information for Multiqc
        help_XICRA.multiqc_help()
        exit()
        
    ## debugging messages
    global Debug
    if (options.debug):
        Debug = True
    else:
        Debug = False
        
    ### set as default paired_end mode
    if (options.single_end):
        options.pair = False
    else:
        options.pair = True
    
    ##################################
    ## start the process
    ##################################
    HCGB_aes.pipeline_header('XICRA')
    HCGB_aes.boxymcboxface("Trimming samples")
    print ("--------- Starting Process ---------")
    HCGB_time.print_time()

    ## absolute path for in & out
    input_dir = os.path.abspath(options.input)
    outdir=""

    ## set mode: project/detached
    if (options.detached):
        outdir = os.path.abspath(options.output_folder)
        options.project = False
    else:
        options.project = True
        outdir = input_dir        
    
    ##################################
    # Trimming adapters
    ##################################
    ## check adapters provided
        ## options.adapters_a
        ## options.adapters_A
        ## options.extra
        
    ## no adapters provided
    if (not options.adapters_a and not options.adapters_A and not options.extra):
        print (colored("** ERROR: No adapter trimming options provided...", 'red'))
        print ("Please provide any option")
        exit()
    
    ## create dictionary with 
    adapters_dict = {}
    if (options.adapters_a):
        adapters_dict['adapter_a'] = options.adapters_a
    
    if (options.adapters_a):
        adapters_dict['adapter_A'] = options.adapters_A
    
    ## set default
    #if not options.min_read_len:
    #    options.min_read_len=15
    
    ## Default parameters:
    trimming_params = {
        "adapter_a": options.adapters_a,
        "adater_A": options.adapters_A,
        "MINLEN": str(options.min_read_len)
        }
   
    ## get files
    print ('+ Getting files from input folder... ')
    print ('+ Mode: fastq.\n+ Extension: ')
    print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
    pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
    
    ## debug message
    if (Debug):
        print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
        print (pd_samples_retrieved)

        print (colored("**DEBUG: adapters_dict **",'yellow'))
        print (adapters_dict)

    ## generate output folder, if necessary
    print ("\n+ Create output folder(s):")
    if not options.project:
        HCGB_files.create_folder(outdir)
    ## for samples
    outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "trimm", options.debug)
    
    ## optimize threads
    name_list = set(pd_samples_retrieved["new_name"].tolist())
    threads_job = HCGB_main.optimize_threads(options.threads, len(name_list)) ## threads optimization
    max_workers_int = int(options.threads/threads_job)

    ## debug message
    if (Debug):
        print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
        print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
        print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))

    print ("+ Trimming adapters for each sample retrieved...")    
    
    # Group dataframe by sample name
    sample_frame = pd_samples_retrieved.groupby(["new_name"])
    
    ## send for each sample
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor:
        commandsSent = { executor.submit(cutadapt_caller.caller, sorted(cluster["sample"].tolist()), 
                                         outdir_dict[name[0]], name[0], threads_job, 
                                         options.min_read_len, Debug, 
                                         adapters_dict, options.extra): name[0] for name, cluster in sample_frame }

        for cmd2 in concurrent.futures.as_completed(commandsSent):
            details = commandsSent[cmd2]
            try:
                data = cmd2.result()
            except Exception as exc:
                print ('***ERROR:')
                print (cmd2)
                print('%r generated an exception: %s' % (details, exc))

    print ("\n\n+ Trimming samples has finished...")
    
    ## HCGB_time.timestamp
    start_time_partial = HCGB_time.timestamp(start_time_total)

    ## get files generated and generate symbolic link
    if not options.project:
        dir_symlinks = HCGB_files.create_subfolder('link_files', outdir)
        files2symbolic = []
        folders = os.listdir(outdir)

        ## debug message
        if (Debug):
            print (colored("**DEBUG: generate symbolic links for each file in " + dir_symlinks + "**", 'yellow'))
        
        for fold in folders:
            if fold.endswith(".log"):
                continue
            else:
                this_folder = outdir + '/' + fold
                subfiles = os.listdir(this_folder)
                for files in subfiles:
                    files_search = re.search(r".*trim_R\d{1}.*", files) ## only paired-end. Todo: single end
                    if files_search:
                        files2symbolic.append(this_folder + '/' + files)
    
        HCGB_files.get_symbolic_link(files2symbolic, dir_symlinks)

    if (options.skip_report):
        print ("+ No report generation...")
    else:
        print ("\n+ Generating a report using MultiQC module.")
        outdir_report = HCGB_files.create_subfolder("report", outdir)
    
        ## call multiQC report module
        givenList = [ v for v in outdir_dict.values() ]
        my_outdir_list = set(givenList)
        
        ## debug message
        if (Debug):
            print (colored("\n**DEBUG: my_outdir_list for multiqc report **", 'yellow'))
            print (my_outdir_list)
            print ("\n")

        trimm_report = HCGB_files.create_subfolder("trim", outdir_report)
        multiQC_report.multiQC_module_call(my_outdir_list, "Cutadapt", trimm_report,"")
        print ('\n+ A summary HTML report of each sample is generated in folder: %s' %trimm_report)
        
        ## QC analysis for trimmed reads
        if (Debug):
            print (colored("** Beginning FAStQC analysis **", 'red'))

        ## HCGB_time.timestamp
        start_time_partial = HCGB_time.timestamp(start_time_partial)

    ## create FASTQC calling for trimmed reads
    pd_samples_retrieved_trimmed = sampleParser.files.get_files(options, input_dir, "trim", ['_trim'], options.debug)
    outdir_dict_trimmed = qc.fastqc(pd_samples_retrieved_trimmed, outdir, options, "trimmed", start_time_partial, Debug)
        
    print ("\n*************** Finish *******************")
    start_time_partial = HCGB_time.timestamp(start_time_total)
    
    ## samples information dictionary
    samples_info = {}
    samples_frame = pd_samples_retrieved.groupby('new_name')
    for name_tuple, grouped in samples_frame:
        name = name_tuple[0]
        samples_info[name] = grouped['sample'].to_list()
    
    samples_info_trimmed = {}
    samples_frame_trimmed = pd_samples_retrieved_trimmed.groupby('new_name')
    for name_tuple, grouped in samples_frame_trimmed:
        name = name_tuple[0]
        samples_info_trimmed[name] = grouped['sample'].to_list()
    
    
    ## dump information and parameters
    info_dir = HCGB_files.create_subfolder("info", outdir)
    print("+ Dumping information and parameters")
    runInfo = { "module":"trim", "time":time.time(),
                "XICRA version":pipeline_version,
                'sample_info': samples_info,
                'trimmed_sample_info': samples_info_trimmed,
                'triming_parameters': trimming_params,
                'outdir_dict': outdir_dict,
                'outdir_dict_trimmed': outdir_dict_trimmed}
    
    HCGB_info.dump_info_run(info_dir, "trim", options, runInfo, options.debug)
    
    ## dump conda details
    HCGB_info.dump_info_conda(info_dir, "trim", package_name="XICRA", debug=options.debug)

    print ("\n+ Exiting trim module.")
    exit()




#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2024 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
"""
Create RNA biotype analysis using STAR and featureCounts
"""
## import useful modules
import os
import time
import shutil
from termcolor import colored

## import my modules
from XICRA import __version__ as pipeline_version
from XICRA.config import set_config
from XICRA.modules import help_XICRA, map
from XICRA.scripts import RNAbiotype, multiQC_report
from XICRA.other_tools import tools

from HCGB import sampleParser
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.info_functions as HCGB_info
import HCGB.functions.main_functions as HCGB_main
import HCGB.functions.system_call_functions as HCGB_system



##############################################
def run_biotype(options):
    """Main function of the module, organizes the biotype analysis

    First, checks if the files needed to do the mapping (and posterior
    classification of the reads) have been provided by the user:
    fasta sequence + annotation or STAR index directory. 

    First, it executes the maping calling map XICRA module.
    Then, it calls RNAbiotype_module_call() to execute featureCounts.
    
    Finally, generate the MultiQC featureCounts report of all samples.
    
    :param options: input parameters introduced by the user. See XICRA biotype -h.

    :returns: None
    """

    ## init time
    start_time_total = time.time()
    
    ##################################
    ### show help messages if desired    
    ##################################
    if (options.help_format):
        ## help_format option
        help_XICRA.help_fastq_format()
    elif (options.help_project):
        ## information for project
        help_XICRA.project_help()
        exit()
    elif (options.help_RNAbiotype):
        ## information for join reads
        RNAbiotype.help_info()
        exit()
        
    ## debugging messages
    global Debug
    if (options.debug):
        Debug = True
    else:
        Debug = False
        
    ### set as default paired_end mode
    if (options.single_end):
        options.pair = False
    else:
        options.pair = True
    
    HCGB_aes.pipeline_header('XICRA')
    HCGB_aes.boxymcboxface("RNA biotype analysis")
    print ("--------- Starting Process ---------")
    HCGB_time.print_time()

    ## absolute path for in & out
    input_dir = os.path.abspath(options.input)
    outdir=""

    ## set mode: project/detached
    if (options.detached):
        outdir = os.path.abspath(options.output_folder)
        options.project = False
    else:
        options.project = True
        outdir = input_dir        
    
    ## get files
    print ('+ Getting files from input folder... ')
    
    ## get files: use joined reads if paired-end data
    if options.pair:
        options.pair = False ## set paired-end to false for further prepocessing
        if options.noTrim:
            print ('+ Mode: fastq.\n+ Extension: ')
            print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
        elif options.noJoin:
            print ('+ Mode: fastq.\n+ Extension: ')
            print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "trim", ["_trim"], options.debug)
        else:
            print ('+ Mode: join.\n+ Extension: ')
            print ("[_joined.fastq]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "join", ['trim_joined.fastq'], options.debug)
    else:
        if options.noTrim:
            print ('+ Mode: fastq.\n+ Extension: ')
            print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
        else:
            print ('+ Mode: trim.\n+ Extension: ')
            print ("[_trim.fastq]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "trim", ['_trim'], options.debug)
    
    ## debug message
    if (Debug):
        print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
        print (pd_samples_retrieved)

    ## generate output folder, if necessary
    print ("\n+ Create output folder(s):")
    if not options.project:
        HCGB_files.create_folder(outdir)

    ## for samples
    mapping_outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "map", options.debug)
    
    ## debug message
    if (Debug):
        print (colored("**DEBUG: mapping_outdir_dict **", 'yellow'))
        print (mapping_outdir_dict)

    # time stamp
    start_time_partial = HCGB_time.timestamp(start_time_total)

    ## optimize threads
    name_list = set(pd_samples_retrieved["new_name"].tolist())
    threads_job = HCGB_main.optimize_threads(options.threads, len(name_list)) ## threads optimization
    max_workers_int = int(options.threads/threads_job)
    
    ## debug message
    if (Debug):
        print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
        print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
        print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))
        
    ##############################################
    ## map Reads
    ##############################################
    
    ## multimapping:
    if options.no_multiMapping:
        multimapping = False
    else:
        multimapping = True
    
    (start_time_partial, mapping_results) = map.mapReads_module_STAR(options, pd_samples_retrieved, mapping_outdir_dict, 
                    options.debug, max_workers_int, threads_job, start_time_partial, outdir, multimapping)

    ## debug message
    if (Debug):
         print (colored("**DEBUG: mapping_results **", 'yellow'))
         print (mapping_results)
    
    # time stamp
    start_time_partial = HCGB_time.timestamp(start_time_partial)
    
    ## for samples
    biotype_outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "biotype", options.debug)
    
    ## debug message
    if (Debug):
        print (colored("**DEBUG: biotype_outdir_dict **", 'yellow'))
        print (biotype_outdir_dict)
    
    ##############################################
    ## feature counts
    ##############################################

    ## get RNAbiotype information
    RNAbiotype.RNAbiotype_module_call(mapping_results, biotype_outdir_dict, options.annotation, 
                                      options.debug, max_workers_int, threads_job, multimapping, options.stranded)

    # time stamp
    start_time_partial = HCGB_time.timestamp(start_time_partial)
    
    if (options.skip_report):
        print ("+ No report generation...")
    else:
        print ("\n+ Generating a report using MultiQC module for featureCount analysis.")
        outdir_report = HCGB_files.create_subfolder("report", outdir)

        ## get subdirs generated and call multiQC report module
        givenList = []
        print ("+ Detail information for each sample could be identified in separate folders:")
        
        ## call multiQC report module
        givenList = [ v for v in biotype_outdir_dict.values() ]
        my_outdir_list = set(givenList)

        ## debug message
        if (Debug):
            print (colored("\n**DEBUG: my_outdir_list for multiqc report **", 'yellow'))
            print (my_outdir_list)
            print ("\n")
        
        featureCount_report = HCGB_files.create_subfolder("featureCount", outdir_report)
        multiQC_report.multiQC_module_call(my_outdir_list, "featureCount", featureCount_report,"-dd 2")
        print ('\n+ A summary HTML report of each sample is generated in folder: %s' %featureCount_report)

        ### Summarizing RNA biotype information
        biotype_report = HCGB_files.create_subfolder("biotype", outdir_report)
        single_files_biotype = HCGB_files.create_subfolder("samples", biotype_report)
        
        ## results
        dict_files = {}
        
        for samples in biotype_outdir_dict:
            featurecount_file = os.path.join(biotype_outdir_dict[samples], 'featureCount.out.tsv')
            if HCGB_files.is_non_zero_file(featurecount_file):
                dict_files[samples] = featurecount_file
            
                ## copy pdf
                pdf_plot = HCGB_main.retrieve_matching_files(biotype_outdir_dict[samples], '.pdf', options.debug)
                if HCGB_files.is_non_zero_file(pdf_plot[0]):
                    shutil.copy(pdf_plot[0], single_files_biotype)
            
        ## collapse all information
        all_data = RNAbiotype.generate_matrix(dict_files)
    
        ## print into excel/csv
        print ('+ Table contains: ', len(all_data), ' entries\n')
        
        ## debugging messages
        if Debug:
            print ("** DEBUG: all_data")
            print (all_data)
        
        ## set abs_csv_outfile to be in report folder
        ## copy or link files for each sample analyzed
        abs_csv_outfile = os.path.join(biotype_report, "summary.csv")
        all_data.to_csv(abs_csv_outfile)
       
        ## create plot: call R and XICRA::stats 
        ## outfile_pdf = os.path.join(biotype_report, "RNAbiotypes_summary.pdf") ## creates automatically the name biotypes-plot.pdf
        
        ## R scripts
        biotype_R_script = tools.R_scripts('plot_RNAbiotype_sum', options.debug)
        rscript = set_config.get_exe("Rscript", options.debug)
        cmd_R_plot = "%s %s -f %s -o %s" %(rscript, biotype_R_script, abs_csv_outfile, biotype_report)
        
        ##
        print ("+ Create summary plot for all samples")
        callCode = HCGB_system.system_call(cmd_R_plot)
    
    ## Create length distribution analysis
    ## get_length_distribution.get_length_dist()
    
    ## Add mirtrace analysis and report
   
    print ("\n*************** Finish *******************")
    start_time_partial = HCGB_time.timestamp(start_time_total)
 
    ## samples information dictionary
    samples_info = {}
    samples_frame = pd_samples_retrieved.groupby('new_name')
    for name_tuple, grouped in samples_frame:
        #name = name_tuple[0]
        samples_info[name_tuple] = grouped['sample'].to_list()
    
    ## dump information and parameters
    info_dir = HCGB_files.create_subfolder("info", outdir)
    print("+ Dumping information and parameters")
    runInfo = { "module":"biotype", "time":time.time(),
                "XICRA version":pipeline_version,
                'mapping_results': mapping_results,
                'outdir_dict': biotype_outdir_dict,
                'featureCounts': dict_files}
    
    HCGB_info.dump_info_run(info_dir, "biotype", options, runInfo, options.debug)
    
    ## dump conda details
    HCGB_info.dump_info_conda(info_dir, "biotype", package_name="XICRA", debug=options.debug)
   
    print ("\n+ Exiting biotype module.")
    return()




#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
"""
This module performs miRNA analysis with three possible tools: 
Miraligner, OptimiR and sRNAbench. It shows results unifiying 
formats by using the miRTop nomenclature.
"""
## import useful modules
import os
import time
import concurrent.futures
import pandas as pd
from termcolor import colored

## import my modules
from HCGB import sampleParser
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.info_functions as HCGB_info
import HCGB.functions.main_functions as HCGB_main

from XICRA import __version__ as pipeline_version
from XICRA.modules import help_XICRA
from XICRA.modules import database
from XICRA.scripts import generate_DE

from XICRA.scripts import mirtop_caller
from XICRA.scripts import sRNAbench_caller
from XICRA.scripts import optimir_caller
from XICRA.scripts import miraligner_caller


##############################################
def run_miRNA(options):
    """Main function of the module, organizes the miRNA analysis

    First, checks if the files needed to do the mapping (and posterior
    classification of the reads) have been provided by the user:
    - miRNA gff3 annotation, hairpin fasta, mature fasta, miRBase str annotation
    If some is missing it will be downloaded from miRBase. 

    Then, it calls miRNA_analysis() for each sample in parallel.
    Gets user software selection: sRNAbench, optimiR, miraligner.
    Standarize results using miRTop.
    Finally, build final matrix comparing all samples.
    
    :param options: input parameters introduced by the user. See XICRA miRNA -h.

    :returns: None
    """

    ## init time
    start_time_total = time.time()

    ##################################
    ### show help messages if desired    
    ##################################
    if (options.help_format):
        ## help_format option
        help_XICRA.help_fastq_format()
    elif (options.help_project):
        ## information for project
        help_XICRA.project_help()
        exit()
    elif (options.help_miRNA):
        ## information for join reads
        help_XICRA.help_miRNA()
        exit()
        
    ## debugging messages
    global Debug
    if (options.debug):
        Debug = True
    else:
        Debug = False
        
    ### set as default paired_end mode
    if (options.single_end):
        options.pair = False
    else:
        options.pair = True
    
    HCGB_aes.pipeline_header('XICRA')
    HCGB_aes.boxymcboxface("miRNA analysis")
    print ("--------- Starting Process ---------")
    HCGB_time.print_time()

    ## absolute path for in & out
    input_dir = os.path.abspath(options.input)
    outdir=""

    ## set mode: project/detached
    if (options.detached):
        outdir = os.path.abspath(options.output_folder)
        options.project = False
    else:
        options.project = True
        outdir = input_dir        
    
    ## user software selection
    print ("+ Software for miRNA analysis selected:")
    print (options.soft_name)
    
    ## get files
    print ('+ Getting files from input folder... ')
    if options.pair:
        options.pair = False ## set paired-end to false for further prepocessing
        if options.noTrim:
            print ('+ Mode: fastq.\n+ Extension: ')
            print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
        else:
            print ('+ Mode: join.\n+ Extension: ')
            print ("[_joined.fastq]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "join", ['_joined.fastq'], options.debug)
    else:
        if options.noTrim:
            print ('+ Mode: fastq.\n+ Extension: ')
            print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
        else:
            print ('+ Mode: join.\n+ Extension: ')
            print ("[_joined.fastq]\n")
            pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "trim", ['_trim'], options.debug)
    
    ## debug message
    if (Debug):
        print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
        print (pd_samples_retrieved)

    ## Additional sRNAbench or miRTop options
    
    ## species
    print ("+ Species provided:", options.species)
    
    ############################################################
    ## Download miRNA information: hairpin, mature, str, gff3
    ############################################################
    if not (options.database):
        install_path =  os.path.dirname(os.path.realpath(__file__))
        options.database = os.path.join(install_path, "db_files") 
    else:
        options.database = os.path.abspath(options.database)
    
    print ("+ Create folder to store results: ", options.database)
    HCGB_files.create_folder(options.database)
    
    ## call database module and return options updated
    options = database.miRNA_db(options)    
    
    ##############################################################
    ## Start the analysis
    ##############################################################
    ## generate output folder, if necessary
    if not options.project:
        print ("\n+ Create output folder(s):")
        HCGB_files.create_folder(outdir)
    
    ## for samples
    outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "miRNA", options.debug)
    
    ## optimize threads
    name_list = set(pd_samples_retrieved["new_name"].tolist())
    threads_job = HCGB_main.optimize_threads(options.threads, len(name_list)) ## threads optimization
    max_workers_int = int(options.threads/threads_job)

    ## debug message
    if (Debug):
        print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
        print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
        print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))

    print ("+ Create a miRNA analysis for each sample retrieved...")    
    
    ## call miRNA_analysis: 
    ## Get user software selection: sRNAbench, optimir, ...
    ## Standarize using miRTop
    
    # Group dataframe by sample name
    sample_frame = pd_samples_retrieved.groupby(["new_name"])
    
    ## send for each sample
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor:
        commandsSent = { executor.submit(miRNA_analysis, sorted(cluster["sample"].tolist()), 
                                         outdir_dict[name[0]], name[0], threads_job, options.miRNA_gff,
                                         options.soft_name, options.matureFasta, options.hairpinFasta, 
                                         options.miRBase_str, options.species, 
                                         options.miRNA_db, Debug): name[0] for name, cluster in sample_frame }

        for cmd2 in concurrent.futures.as_completed(commandsSent):
            details = commandsSent[cmd2]
            try:
                data = cmd2.result()
            except Exception as exc:
                print ('***ERROR:')
                print (cmd2)
                print('%r generated an exception: %s' % (details, exc))

    print ("\n\n+ miRNA analysis is finished...")
    print ("+ Let's summarize all results...")
    
    ## outdir
    outdir_report = HCGB_files.create_subfolder("report", outdir)
    expression_folder = HCGB_files.create_subfolder("miRNA", outdir_report)

    ## Get results files generated
    
    ## dictionary results
    results_SampleParser = sampleParser.files.get_files(options, input_dir, "miRNA", ["mirtop.tsv"], options.debug)
    results_df = pd.DataFrame(columns=("name", "soft", "filename"))
    results_df['name'] = results_SampleParser['name']
    results_df['soft'] = results_SampleParser['ext']
    results_df['filename'] = results_SampleParser['sample']

    ## debugging messages
    if options.debug:
        print (results_df)
    
    ## merge all parse gtf files created
    print ("+ Summarize miRNA analysis for all samples...")
    generate_DE.generate_DE(results_df, options.debug, expression_folder)

    print ("\n*************** Finish *******************")
    HCGB_time.timestamp(start_time_total)
    
    ## samples information dictionary
    samples_info = {}
    samples_frame = pd_samples_retrieved.groupby('new_name')
    for name_tuple, grouped in samples_frame:
        #name = name_tuple[0]
        samples_info[name_tuple] = grouped['sample'].to_list()
    
    ## dump information and parameters
    info_dir = HCGB_files.create_subfolder("info", outdir)
    print("+ Dumping information and parameters")
    runInfo = { "module":"mirna", "time":time.time(),
                "XICRA version":pipeline_version,
                'outdir_dict': outdir_dict,
                'expression_folder':expression_folder,
                'results_df': results_df.to_dict()}
    
    HCGB_info.dump_info_run(info_dir, "mirna", options, runInfo, options.debug)
    
    ## dump conda details
    HCGB_info.dump_info_conda(info_dir, "mirna", package_name="XICRA", debug=options.debug)
    
    print ("\n+ Exiting miRNA module.")
    return()

###############
def miRNA_analysis(reads, folder, name, threads, miRNA_gff, soft_list, 
                   matureFasta, hairpinFasta, miRBase_str, species, database, Debug):
    """Passes the sample information to be analyzed by the selected softwares

    Checks the introduced softwares and calls them one by one to analyze the given
    sample. It stores their results in separated folders, which are called as the 
    used software. Finally, it unifies each software output in miRTop format.
    
    :param reads: file with sample reads
    :param folder: output folder
    :param name: sample name
    :param threads: selected threads (by defoult 2)
    :param miRNA_gff: miRNA gff3 annotation file
    :param soft_list: list with the selected softwares 
    :param matureFasta: mature fasta file 
    :param hairpinFasta: hairpin fasta file
    :param miRBase_str: miRBase str annotation 
    :param species: species tag ID. Default: hsa (Homo sapiens)
    :param database: path to store miRNA annotation files downloaded
    :param Debug: display complete log.


    :returns: None
    """
    
    for soft in soft_list:
        if (soft == "sRNAbench"):
            ## create sRNAbench
            sRNAbench_folder = HCGB_files.create_subfolder('sRNAbench', folder)
            code_success = sRNAbench_caller.sRNAbench_caller(reads, sRNAbench_folder, name, threads, species, Debug) ## Any additional sRNAbench parameter?
                
            if not code_success:
                print ('** miRTop would not be executed for sample %s...' %name)
                return ()
            
            ## create folder for sRNAbench results
            miRTop_folder = HCGB_files.create_subfolder("sRNAbench_miRTop", folder)
            mirtop_caller.miRTop_caller(sRNAbench_folder, miRTop_folder, name, threads, miRNA_gff, hairpinFasta, 'sRNAbench', species, Debug)
            
        ###
        if (soft == "optimir"):
            ## create OptimiR analysis
            optimir_folder = HCGB_files.create_subfolder('OptimiR', folder)
            code_success = optimir_caller.optimir_caller(reads, optimir_folder, name, threads, matureFasta, hairpinFasta, miRNA_gff, species, Debug) ## Any additional sRNAbench parameter?
            
            ## create folder for Optimir results
            miRTop_folder = HCGB_files.create_subfolder("OptimiR_miRTop", folder)
            mirtop_caller.miRTop_caller(optimir_folder, miRTop_folder, name, threads, miRNA_gff, hairpinFasta, 'optimir', species, Debug)
            
        ###
        if (soft == "miraligner"):
            
            ## create OptimiR analysis
            miraligner_folder = HCGB_files.create_subfolder('miraligner', folder)
            code_success = miraligner_caller.miraligner_caller(reads, miraligner_folder, name, threads, database, species, Debug) 
            
            ## create folder for Optimir results
            miRTop_folder = HCGB_files.create_subfolder("miraligner_miRTop", folder)
            mirtop_caller.miRTop_caller(miraligner_folder, miRTop_folder, name, threads, miRNA_gff, hairpinFasta, 'seqbuster', species, Debug)
            



from pkg_resources import get_distribution
try:
	__version__ = get_distribution('XICRA').version
except:
	try:
		__version__ = pkg_resources.resorce_filename('XICRA', VERSION)
	except:
		__version__ = 'local'

__all__ = [
	'modules',
	'scripts',
	'config',
	'other_tools'
]

from XICRA import *




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain        ##
##########################################################
"""
Get fastq files and prepare them for other purposes. 
"""

## import useful modules
import time
import os
from io import open
import shutil
import pandas as pd
from termcolor import colored
from XICRA import __version__ as pipeline_version

## import my modules
from XICRA.modules import help_XICRA
#from HCGB import functions
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.info_functions as HCGB_info

from HCGB import sampleParser

################################
def run_prep(options):
    """
    Main function of the prep module.
    
    This module prepares fastq files for later usage. It initially checks the length
    of the name and advises the user to rename samples if exceeded. 
    
    This module allows to user to copy files into the project folder initiate or only link using
    a symbolic link to avoid duplicated raw data. 
    
    The input files will be organized in folders. One folder per sample. 

    :param options: input parameters introduced by the user. See XICRA prep -h.

    :returns: None
    """
    
    ## help_format option
    if (options.help_format):
        help_XICRA.help_fastq_format()
        exit()
        
    HCGB_aes.pipeline_header('XICRA')
    HCGB_aes.boxymcboxface("Preparing samples")
    print ("--------- Starting Process ---------")
    HCGB_time.print_time()
    
    ## init time
    start_time_total = time.time()
    
    ## absolute path for in & out
    input_dir = os.path.abspath(options.input)
    outdir = os.path.abspath(options.output_folder)

    ### set as default paired_end mode
    if (options.single_end):
        options.pair = False
    else:
        options.pair = True

    ## output folder    
    print ("\n+ Create output folder(s):")
    HCGB_files.create_folder(outdir)

    ## project
    if options.detached:
        options.project = False
    else:
        options.project = True
    
    ## default options
    if not options.include_lane:
        options.include_lane = False
    
    if not options.include_all:
        options.include_all = False

    ### info
    final_dir = ""
    if (options.project):
        print ("+ Generate a directory containing information within the project folder provided")
        final_dir = HCGB_files.create_subfolder("info", outdir)
    else:
        final_dir = outdir
    
    ## get files
    print ()
    HCGB_aes.print_sepLine("-",50, False)
    print ('+ Getting files from input folder... ')
    print ('+ Mode: fastq.\n+ Extension: ')
    print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
    
    pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)
        
    ## Information returned in pd_samples_retrieved
    ### sample, dirname, name, name_len, lane, read_pair, lane_file, ext, gz
    
    if options.debug:
        print (colored("** DEBUG: pd_samples_retrieved", 'yellow'))
        print (pd_samples_retrieved)
        #functions.print_all_pandaDF(pd_samples_retrieved)
    
    ## time stamp
    start_time_partial = HCGB_time.timestamp(start_time_total)
    
    ## check character limitation
    list_lengths = pd_samples_retrieved.loc[:,'name_len'].to_list()
    if any(i > 25 for i in list_lengths):
        print (colored("\t ** Name lengths exceeds the 25 character limitation...", 'yellow'))
        if not (options.rename):
            print (colored("** ERROR: Rename files or provide --rename option...", 'red'))
            exit()

    ### rename files 
    if (options.rename):
        options.rename = os.path.abspath(options.rename)
        if not HCGB_files.is_non_zero_file(options.rename):
            print (colored("** ERROR: File provided with rename information is not readable.", 'red'))
            print (options.rename)
            exit()
            
        names_retrieved = pd.read_csv(options.rename, sep=',', 
                                    index_col=0, squeeze=True, 
                                    header=None).to_dict() ## read csv to dictionary
        if (options.debug):
            print (colored('** DEBUG: names_retrieved', 'yellow'))
            print (names_retrieved)
            
        ## TODO: check integrity of new names and special characters
    
        ## print to a file
        timestamp = HCGB_time.create_human_timestamp()
        rename_details = final_dir + '/' + timestamp + '_prep_renameDetails.txt'
        rename_details_hd = open(rename_details, 'w')
    
        ## rename files         
        for index, row in pd_samples_retrieved.iterrows():
            if (row['gz']):
                extension_string = row['ext'] + row['gz']
            else:
                extension_string = row['ext']
            
            if options.single_end:
                renamed = names_retrieved[row['name']] + '.' + extension_string
            else:
                renamed = names_retrieved[row['name']] + '_' + row['read_pair'] + '.' + extension_string
            
            ## modify frame
            pd_samples_retrieved.loc[index, 'new_file'] = renamed
            pd_samples_retrieved.loc[index, 'new_name'] = names_retrieved[row['name']]
            ## save in file
            string = row['sample'] + '\t' + renamed + '\n'
            rename_details_hd.write(string)
            
            if (options.debug):
                print (colored('** DEBUG: rename', 'yellow'))
                print ("Original: ", row['name'])
                print ("Renamed: ", names_retrieved[row['name']])
                print ("File:", renamed)
        
        rename_details_hd.close()    

        ##elif (options.single_end): It should work for both
        print ("+ Sample files have been renamed...")
    else:
        pd_samples_retrieved['new_file'] = pd_samples_retrieved['file']
 
    ## create outdir for each sample
    outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, "raw", options.debug, groupby_col="new_name")    
        
    ## merge option
    if (options.merge_Reads):
        print ("+ Sample files will be merged...")
        ## TODO: check when rename option provided
        pd_samples_merged = sampleParser.merge.one_file_per_sample(
            pd_samples_retrieved, outdir_dict, options.threads,    
            final_dir, options.debug)
        
        if (options.rename):
            print ("+ Merge files have been renamed...")
        else:
            print ("+ Sample files have been merged...")
        
        ## process is finished here
        print ("\n*************** Finish *******************")
        start_time_partial = HCGB_time.timestamp(start_time_total)
    
        print ("+ Exiting prep module.")
        exit()
    
    ## debugging messages
    if (options.debug):
        print (colored("** DEBUG: pd_samples_retrieved", 'yellow'))
        #functions.print_all_pandaDF(pd_samples_retrieved)
        print (pd_samples_retrieved)
        print (colored("** DEBUG: outdir_dict", 'yellow'))
        print (outdir_dict)
    
    ## copy or create symbolic link for files
    if (options.copy_reads):
        print ("+ Sample files will be copied...")
        ## print to a file
        timestamp = HCGB_time.create_human_timestamp()
        copy_details = final_dir + '/' + timestamp + '_prep_copyDetails.txt'
        copy_details_hd = open(copy_details, 'w')
    else:
        print ("+ Sample files will be linked...")    
    
    list_reads = []
    for index, row in pd_samples_retrieved.iterrows():
        if (options.copy_reads):
            ## TODO: debug & set threads to copy faster
            shutil.copy(row['sample'], os.path.join(outdir_dict[row['new_name']], row['new_file'] ))            
            string = row['sample'] + '\t' + os.path.join(outdir_dict[row['new_name']], row['new_file']) + '\n'
            copy_details_hd.write(string)            
        else:
            list_reads.append(row['new_file'])
            
            if options.project:
                HCGB_files.get_symbolic_link_file(row['sample'], 
                                                 os.path.join(outdir_dict[row['new_name']], row['new_file']))

    if (options.copy_reads):
        print ("+ Sample files have been copied...")
        copy_details_hd.close()
    else:
        if not options.project:
            HCGB_files.get_symbolic_link(list_reads, outdir)
    
    print ("\n*************** Finish *******************")
    HCGB_time.timestamp(start_time_total)


    ## samples information dictionary
    samples_info = {}
    samples_frame = pd_samples_retrieved.groupby('name')
    for name, grouped in samples_frame:
        samples_info[name] = grouped['sample'].to_list()
    
    ## dump information and parameters
    info_dir = HCGB_files.create_subfolder("info", outdir)
    print("+ Dumping information and parameters")
    runInfo = { "module":"prep",  "time":time.time(),
                "XICRA version":pipeline_version,
                'sample_info': samples_info}
    
    HCGB_info.dump_info_run(info_dir, "prep", options, runInfo, options.debug)
    
    ## dump conda details
    HCGB_info.dump_info_conda(info_dir, "prep", package_name="XICRA", debug=options.debug)

    print ("+ Exiting prep module.")
    return()



#!/usr/bin/env python3
############################################################
## Jose F. Sanchez, Marta Lopez & Lauro Sumoy             ##
## Copyright (C) 2019-2021 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################

from HCGB.functions import aesthetics_functions
import HCGB

"""
This module downloads data for genome annotation, miRNA, tRNA and piRNA analysis: 
"""

## import useful modules
import os
import sys
import re
from io import open
from termcolor import colored

## import my modules
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.main_functions as HCGB_main
import HCGB.functions.time_functions as HCGB_time


##############################################
def run_db(options):
    print()
    
    
##############################################
def miRNA_db(options):
    
    ############################################################
    ## miRNA information: hairpin, mature, str, gff3 from miRBase
    ############################################################
    
    Debug = options.debug
    
    options.miRNA_db = os.path.join(options.database, "miRNA_db")
    HCGB_files.create_folder(options.miRNA_db)
    
    ## First check if already provided files
    list_files = HCGB_main.get_fullpath_list(options.miRNA_db, options.debug)
    
    ## Check for files from miRBase:
    miRBase_files = ["hsa.gff3", "hairpin.fa", "mature.fa", "miRNA.str"]
    miRBase_files_dict = {}
    download_data=False
    
    ## get file names
    list_names = list(map(os.path.basename, list_files))
    
    if Debug:
        aesthetics_functions.debug_message("list_files: " + str(list_files))
        aesthetics_functions.debug_message("list_names: " + str(list_names))
    
    for file_req in miRBase_files:
        if Debug:
            aesthetics_functions.debug_message("file_req: " + file_req)
            
        if file_req not in list_names:
            download_data=True
            miRBase_files_dict[file_req] = ""
        else:
            file_retrieved = HCGB_main.retrieve_matching_files(options.miRNA_db, file_req, options.debug, starts=False)
            if HCGB_files.is_non_zero_file(file_retrieved[0]):
                miRBase_files_dict[file_req] = file_retrieved[0]
            else:
                miRBase_files_dict[file_req] = ""

    ## If missing, download them, if all files ok, return!
    if Debug:
        aesthetics_functions.debug_message("miRBase_files_dict: ")
        print(miRBase_files_dict)
    
    ## miRBase ftp site
    ftp_site = "https://www.mirbase.org/ftp/CURRENT/"
    
    ## -------------------------------
    ## miRNA_gff: can be set as automatic to download from miRBase
    ## -------------------------------
    if not miRBase_files_dict['hsa.gff3']:
        print ("+ File miRNA gff3 annotation")
        if Debug:
            print (colored("\t** ATTENTION: No miRNA gff file provided", 'yellow'))     
        print (colored("\t** Download it form miRBase", 'green'))
        file_name = options.species + ".gff3"
        ftp_site1 = "https://www.mirbase.org/ftp/CURRENT/genomes/" 
        options.miRNA_gff = HCGB_main.urllib_request(options.miRNA_db, ftp_site1, file_name, Debug)
        
    else:
        if (options.miRNA_gff):
            print ("\t+ miRNA gff file provided")
            options.miRNA_gff = os.path.abspath(options.miRNA_gff)
        else:
            print ("\t+ miRNA gff file available")
            options.miRNA_gff = miRBase_files_dict['hsa.gff3']

    ## -------------------------------
    ## hairpin: can be set as automatic to download from miRBase
    ## -------------------------------
    if not miRBase_files_dict['hairpin.fa']:
        print ("+ File hairpin fasta")
        if Debug:
            print (colored("\t** ATTENTION: No hairpin fasta file provided", 'yellow'))        
        print (colored("\t** Download it form miRBase", 'green'))
        options.hairpinFasta = HCGB_main.urllib_request(options.miRNA_db, ftp_site, "hairpin.fa.gz", Debug)
        
    else:
        if (options.hairpinFasta):
            print ("\t+ hairpin fasta file provided")
            options.hairpinFasta = os.path.abspath(options.hairpinFasta)
        else:
            print ("\t+ hairpin fasta file available")
            options.hairpinFasta = miRBase_files_dict['hairpin.fa']
            
   
    ## -------------------------------
    ## mature: can be set as automatic to download from miRBase
    ## -------------------------------
    if not miRBase_files_dict['mature.fa']:
        print ("+ File mature fasta")
        if Debug:
            print (colored("\t** ATTENTION: No mature miRNA fasta file provided", 'yellow'))        
        print (colored("\t** Download it form miRBase", 'green'))
        options.matureFasta = HCGB_main.urllib_request(options.miRNA_db, ftp_site, "mature.fa.gz", Debug)

    else:
        if (options.matureFasta):
            print ("\t+ mature fasta file provided")
            options.matureFasta = os.path.abspath(options.matureFasta)
        else:
            print ("\t+ miRNA gff file available")
            options.matureFasta = miRBase_files_dict['mature.fa']
        
    ## -------------------------------
    ## miRBase str: can be set as automatic to download from miRBase
    ## -------------------------------
    if not miRBase_files_dict['miRNA.str']:
        print ("+ File miRBase str annotation")
        if Debug:
            print (colored("\t** ATTENTION: No miRBase_str file provided", 'yellow'))        
        print (colored("\t** Download it form miRBase", 'green'))
        options.miRBase_str = HCGB_main.urllib_request(options.miRNA_db, ftp_site, "miRNA.str.gz", Debug)
        ## extract
        
    else:
        if (options.miRBase_str):
            print ("\t+ miRBase_str file provided")
            options.miRBase_str = os.path.abspath(options.miRBase_str)
        else:
            print ("\t+ miRBase str file available")
            options.miRBase_str = miRBase_files_dict['miRNA.str']
        
    ## -------------------------------
    return (options)    

##############################################
def tRNA_db(database, debug):
    
    ## Mintmap automatically installs and generates database in python path
    ## ../anaconda3/envs/XICRA/lib/python3.7/site-packages/mintmap/mappingbundle/v2/
    
    return True

    # We can try to check if it correctly generated...
    ## TODO
    print ("+ Create folder to store several databases: ", database)
    HCGB_files.create_folder(database)
    
    tRNA_db = os.path.join(database, "tRNA_db")
    if os.path.isdir(tRNA_db):
        code = check_tRNA_db(tRNA_db)
        if code:
            return tRNA_db
    
    ## If missing, download them, if all files ok, return!
    print ("+ Create folder to store tRNA information: ", tRNA_db)
    HCGB_files.create_folder(tRNA_db)
    
    
    
    ## folder and database generated   
    return (tRNA_db)

##############################################
def check_tRNA_db(path2test, debug=False):
    ## Check for
    # LookupTable.tRFs.MINTmap_v2.txt
    # OtherAnnotations.MINTmap_v2.txt
    # tRNAspace.Spliced.Sequences.With49ntFlank.MINTmap_v2.fa
    # tables.cfg
    
    ## First check if already provided files
    list_files = HCGB_main.get_fullpath_list(path2test, debug)
    
    print(list_files)
    return True

##############################################
def piRNA_db(database, piRNA_db, debug):
    
    
    print ("+ Create folder to store several databases: ", database)
    HCGB_files.create_folder(database)
    
    if not piRNA_db:
        piRNA_db = os.path.join(database, "piRNA_db")
    
    print ("+ Create folder to store piRNA information: ", piRNA_db, debug)
    HCGB_files.create_folder(piRNA_db)
    
    ## First check if already provided files
    list_files = HCGB_main.get_fullpath_list(piRNA_db, debug)
    
    ## Check for files:
    
    ## piRBase
    # hsa.v3.0.fa
    # hsa.gold.fa
    # hsa.hg19.bed
    
    ## If missing, download them, if all files ok, return!
    ## TODO
    
    return()
    
    ## Download information
    
    ## piRBase
    ## http://bigdata.ibp.ac.cn/piRBase/about.php
    piRBase_hsa = "http://bigdata.ibp.ac.cn/piRBase/download/v3.0/fasta/hsa.v3.0.fa.gz"
    piRBase_hsa_gold = "http://bigdata.ibp.ac.cn/piRBase/download/v3.0/fasta/hsa.gold.fa.gz"
    piRBase_hsa_bed = "http://bigdata.ibp.ac.cn/piRBase/download/v3.0/bed/hsa.hg19.bed.gz"
    
    
    ## some other
    
    ## http://hammelllab.labsites.cshl.edu/software/
    ## TEsmall genome annotation for Homo sapiens hg38
    piRNA="http://labshare.cshl.edu/shares/mhammelllab/www-data/TEsmall/hg38.tar.gz" 
    
    ## DFam database
    ## TSV list of all matches found in the given assembly that score above the GA threshold.
    DFam_db = "https://www.dfam.org/releases/Dfam_3.5/annotations/hg38/hg38.hits.gz"
    
    ## Non-redundant
    ## TSV list of all non-redundant matches found in the given assembly and that score above the GA threshold.
    DFam_db_nrph = "https://www.dfam.org/releases/Dfam_3.5/annotations/hg38/hg38.nrph.hits.gz" 
    
    ## release note DFam: https://www.dfam.org/releases/Dfam_3.5/relnotes.txt
    
    
    return()

####################################################
def piRNA_info(database_folder, species_name="hsa", Debug=False):
    """
    Given a folder it looks for piRNA_db subfolder and gets files
    """
    
    ## to return
    annot_info = {}
    annot_info["piRBase"] = {}
    annot_info["TEsmall_db"] = {}
    
    ## See file in XICRA/config/database/database_structure.txt for tree structure details of this folder
    
    database_folder = os.path.abspath(database_folder)
    piRNA_db = os.path.join( database_folder, "piRNA_db")
    
    ## 
    list_files_piRNA = HCGB_main.get_fullpath_list(piRNA_db, Debug)
    
    ## debugging messages
    if Debug:
        HCGB_aes.debug_message("piRNA_db: " + piRNA_db, "yellow")
        HCGB_aes.debug_message("list_files_piRNA: " + str(list_files_piRNA), "yellow")
        print()
    
    ## loop to retrieve all files required
    for piRNA_files in list_files_piRNA:
        ## debugging messages
        #if Debug:
        #    HCGB_aes.debug_message("piRNA_file: " + piRNA_files, "yellow")
        
        ## get file name
        file_name = os.path.basename(piRNA_files)
        
        ## get gold piRNA for given species
        species_gold_file = str(species_name) + ".gold.fa"
        ## might be some species
        if file_name == species_gold_file:
            if HCGB_files.is_non_zero_file(piRNA_files):
                annot_info["piRBase"]["gold_piRNA"] = piRNA_files
    
        ## get all bed annotation for given species
        if piRNA_files.endswith(".bed"):
            ## store bed files from piRBase
            bed_piRBase = re.search(".*piRBase.*", piRNA_files) ## piRBase db files
            if bed_piRBase:
                if file_name.startswith(species_name) and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["piRBase"]["bed_file"] = piRNA_files
                
            ## store bed files from TEsmall_db
            bed_TEsmall_db = re.search(".*TEsmall_db.*", piRNA_files) ## TEsmall_db db files
            if bed_TEsmall_db:
                if species_name=="hsa":
                    
                    ## coding genes
                    if file_name.startswith("exon") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["exon"] = piRNA_files
                    if file_name.startswith("intron") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["intron"] = piRNA_files
                    
                    ## miRNA
                    if file_name.startswith("miRNA") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["miRNA"] = piRNA_files
                    if file_name.startswith("hairpin") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["hairpin"] = piRNA_files
                    
                    ## TE elements
                    if file_name.startswith("TE") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["TE"] = piRNA_files
                    
                    ## structural RNA
                    if file_name.startswith("structural_RNA") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["structural_RNA"] = piRNA_files
        
                    ## piRNA cluster
                    if file_name.startswith("piRNA_cluster") and HCGB_files.is_non_zero_file(piRNA_files):
                        annot_info["TEsmall_db"]["piRNA_cluster"] = piRNA_files

    ## debugging messages
    if Debug:
        HCGB_aes.debug_message("annot_info: " + str(annot_info), "yellow")
        print()
    
    ## concatenate all ncRNA annotation, except piRNA, and coding genes to substract later
    ncRNA_bed = os.path.join(piRNA_db, "ncRNA_merged.bed")
    ncRNA_timestamp = os.path.join(piRNA_db, ".merged_ncRNA_success")
    
    ## check if previously trimmed and succeeded
    if os.path.isfile(ncRNA_timestamp):
        stamp = HCGB_time.read_time_stamp(ncRNA_timestamp)
        print (colored("\tA previous command generated results on: %s [%s]" %(stamp, 'merged ncRNA'), 'yellow'))
    
    else:
        ## merge
        files_to_merge = [
            annot_info["TEsmall_db"]["exon"], annot_info["TEsmall_db"]["intron"], 
            annot_info["TEsmall_db"]["structural_RNA"], annot_info["TEsmall_db"]["hairpin"], 
            annot_info["TEsmall_db"]["miRNA"]]
        
        ## 
        original_stdout = sys.stdout # Save a reference to the original standard output
        with open(ncRNA_bed, 'w') as f_out:
            sys.stdout = f_out # Change the standard output to the file we created.
            for f_in in files_to_merge:
                for line in open(f_in):
                    line=line.rstrip().split("\t")
                    line[0] = line[0].split("chr")[1]                    
                    print("\t".join(line))

        sys.stdout = original_stdout
        f_out.close()
        HCGB.functions.time_functions.print_time_stamp(ncRNA_timestamp)

    ## save into dictionary info
    annot_info['general'] = {}
    annot_info['general']['ncRNA'] = ncRNA_bed
    annot_info['general']['timestamp'] = ncRNA_timestamp
    
    ## ---------------------------------
    ## return when finished
    return(annot_info)


            



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain        ##
##########################################################
"""
Calls multiQC to generate HTML statistics reports.
"""
## useful imports
import os
import io
import sys
from io import open
from sys import argv
from termcolor import colored

## import my modules
from HCGB import functions
from XICRA.config import set_config

############
def multiQC_module_call(givenList, name, path, option):
    """
    Prepares files for multiQC report generation.
    
    :param givenList: List of folder to search for multiQC report.
    :param name: Name to include in the html report.
    :param path: Absolute path for the output folder.
    :param option: Some options to provide to multiQC_call.
    
    :type givenList: list
    :type name: string
    :type path: string
    :type option: string
    
    .. seealso:: This function depends on other XICRA functions called:
    
        - :func:`XICRA.scripts.functions.printList2file`
        
        - :func:`XICRA.scripts.multiQC_report.multiQC_call`
    
    """
    pathFile = path + '/' + 'samples.txt'
    functions.main_functions.printList2file(pathFile, givenList)
    multiQC_call(pathFile, name, path, option)    
    
############
def multiQC_call(pathFile, name, folder, option):
    """
    multiQC_ report generation call.
    
    :param pathFile: File containing list of files to include in report.
    :param name: Name to include in the html report.
    :param folder: Absolute path for the output folder.
    :param option: Options to provide to multiQC call.
    
    :type pathFile: string
    :type name: string 
    :type folder: string 
    :type option: string
    
    :returns: :func:`XICRA.scripts.functions.system_call_functions.system_call` output (OK/FALSE)
        
    .. seealso:: This function depends on other XICRA functions called:
    
        - :func:`XICRA.scripts.functions.system_call_functions.system_call`
    
    """
    multiqc_bin = set_config.get_exe("multiqc")
    ## set options for call
    cmd = "%s --force -o %s -n %s -l %s -p -i 'MultiQC report' -b 'HTML report generated for multiple samples and steps' %s" %(multiqc_bin, folder, name, pathFile, option)
    
    ## if a report was previously generated in the folder 
    ## force to delete and generate a new one
    return(functions.system_call_functions.system_call(cmd))



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez, Marta Lopez and Lauro Sumoy         ##
## Copyright (C) 2019-2024 Lauro Sumoy Lab, IGTP, Spain ##
##########################################################
"""
Creates Quality check sequence adapters within fastq reads.
"""
## import useful modules
import os
import time
import concurrent.futures
from termcolor import colored


## import my modules
from XICRA import __version__ as pipeline_version
from XICRA.scripts import multiQC_report
from XICRA.scripts import fastqc_caller
from XICRA.modules import help_XICRA
from HCGB import sampleParser
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.info_functions as HCGB_info
import HCGB.functions.main_functions as HCGB_main

##############################################
def run_QC(options):
    
    ## init time
    start_time_total = time.time()

    ##################################
    ### show help messages if desired    
    ##################################
    if (options.help_format):
        ## help_format option
        sampleParser.help_format()
        exit()
    elif (options.help_project):
        ## information for project
        help_XICRA.project_help()
        exit()
    elif (options.help_multiqc):
        ## information for Multiqc
        multiQC_report.multiqc_help()
        exit()
        
    ## debugging messages
    global Debug
    if (options.debug):
        Debug = True
    else:
        Debug = False    
    
    ### set as default paired_end mode
    if (options.single_end):
        options.pair = False
    else:
        options.pair = True
    
    ## set main header
    HCGB_aes.pipeline_header('XICRA')
    HCGB_aes.boxymcboxface("Quality check")
    print ("--------- Starting Process ---------")
    HCGB_time.print_time()

    ## absolute path for in & out
    input_dir = os.path.abspath(options.input)
    outdir=""

    ## Project mode as default
    if (options.detached):
        options.project = False
        outdir = os.path.abspath(options.output_folder)
    else:
        options.project = True
        outdir = input_dir        
    
    HCGB_aes.boxymcboxface("FASTQC Quality check for samples")
    
    ## get files
    print ('+ Getting files from input folder... ')
    print ('+ Mode: fastq.\n+ Extension: ')
    print ("[ fastq, fq, fastq.gz, fq.gz ]\n")
    pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ["fastq", "fq", "fastq.gz", "fq.gz"], options.debug)

    ## create FASTQC call
    outdir_dict = fastqc(pd_samples_retrieved, outdir, options, "", start_time_total, Debug)

    print ("\n*************** Finish *******************")
    HCGB_time.timestamp(start_time_total)

    ## samples information dictionary
    samples_info = {}
    samples_frame = pd_samples_retrieved.groupby('new_name')
    for name_tuple, grouped in samples_frame:
        #name = name_tuple[0]
        samples_info[name_tuple] = grouped['sample'].to_list()
    
    ## dump information and parameters
    info_dir = HCGB_files.create_subfolder("info", outdir)
    print("+ Dumping information and parameters")
    runInfo = { "module":"qc", "time":time.time(),
                "XICRA version":pipeline_version,
                'sample_info': samples_info,
                'outdir_dict': outdir_dict}
    
    HCGB_info.dump_info_run(info_dir, "qc", options, runInfo, options.debug)
    
    ## dump conda details
    HCGB_info.dump_info_conda(info_dir, "qc", package_name="XICRA", debug=options.debug)


    print ("+ Exiting qc module.")
    exit()

#######################
def fastqc(pd_samples_retrieved, outdir, options, name_analysis, time_stamp, Debug):
    '''
    This is a main function to prepare data to call FASTQC.
    
    :param pd_samples_retrieved
    :param outdir
    :param options
    :param name_analysis
    :param Debug
    
    :type pd_samples_retrieved
    :type outdir
    :type options
    :type name_analysis
    :type Debug
    
    '''
    
    ## debug message
    if (Debug):
        print (colored("\n**DEBUG: pd_samples_retrieve **", 'yellow'))
        print (pd_samples_retrieved)
        print ("\n")

    ## generate output folder, if necessary
    print ("\n+ Create output folder(s):")
    
    ## if not project, outdir contains the dir to put output
    ## in this case, in some other cases might not occur    
    if not options.project:
        HCGB_files.create_folder(outdir)
        
    ## folder name
    if (name_analysis):
        fold_name = "fastqc_" + name_analysis
    else:
        fold_name = "fastqc"

    ## create output dirs for each sample    
    outdir_dict = HCGB_files.outdir_project(outdir, options.project, pd_samples_retrieved, fold_name, options.debug, groupby_col="new_name")
    
    print ("+ Checking quality for each sample retrieved...")
    start_time_partial = time_stamp
    
    # Group dataframe by sample name
    sample_frame = pd_samples_retrieved.groupby(["name"])

    ## optimize threads
    name_list = set(pd_samples_retrieved["name"].tolist())
    threads_job = HCGB_main.optimize_threads(options.threads, len(name_list)) ## threads optimization
    max_workers_int = int(options.threads/threads_job)

    ## debug message
    if (Debug):
        HCGB_aes.debug_message("options.threads: " + str(options.threads), "yellow")
        HCGB_aes.debug_message("max_workers: " + str(max_workers_int), "yellow")
        HCGB_aes.debug_message("threads_job: " + str(threads_job), "yellow")

    ## send for each sample
    print ("+ Calling fastqc for samples...")    
    with concurrent.futures.ThreadPoolExecutor(max_workers=int(max_workers_int)) as executor:
        commandsSent = { executor.submit(fastqc_caller.run_module_fastqc, 
                                         outdir_dict[name[0]], sorted( cluster["sample"].tolist() ), 
                                         name[0], threads_job): name[0] for name, cluster in sample_frame }
        
        for cmd2 in concurrent.futures.as_completed(commandsSent):
            details = commandsSent[cmd2]
            try:
                data = cmd2.result()
            except Exception as exc:
                print ('***ERROR:')
                print (cmd2)
                print('%r generated an exception: %s' % (details, exc))

    print ("+ FASTQC for samples has finished...")    
    
    ## functions.timestamp
    start_time_partial = HCGB_time.timestamp(start_time_partial)

    if (options.skip_report):
        print ("+ No report generation...")
    else:
        print ("\n+ Generating a report using MultiQC module.")
        outdir_report = HCGB_files.create_subfolder("report", outdir)

        ## get subdirs generated and call multiQC report module
        givenList = []
        print ("+ Detail information for each sample could be identified in separate folders:")
        
        ## call multiQC report module
        givenList = [ v for v in outdir_dict.values() ]
        my_outdir_list = set(givenList)

        ## debug message
        if (Debug):
            print (colored("\n**DEBUG: my_outdir_list for multiqc report **", 'yellow'))
            print (my_outdir_list)
            print ("\n")
        
        fastqc_report = HCGB_files.create_subfolder("FASTQC", outdir_report)
        fastqc_final_report = HCGB_files.create_subfolder(fold_name, fastqc_report)
        multiQC_report.multiQC_module_call(my_outdir_list, "FASTQC", fastqc_final_report,"")
        print ('\n+ A summary HTML report of each sample is generated in folder: %s' %fastqc_final_report)

    return(outdir_dict)



#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2021 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
## useful imports
import time
import io
import os
import re
import sys
import csv
from sys import argv
import subprocess
import argparse
from termcolor import colored
import pandas as pd
from collections import defaultdict

import numpy as np
from operator import itemgetter
import sys, csv, os, argparse

from XICRA.config import set_config
from XICRA.scripts import bedtools_caller
from XICRA.scripts import BAMtoPILFER
from XICRA.modules import database

import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.system_call_functions as HCGB_sys
import HCGB.functions.fasta_functions as HCGB_fasta
import HCGB.functions.time_functions as HCGB_time
import HCGB.format_conversion

##########################################################3
def pilfer_merge_samples_call(pilfer_results, pilfer_folder, debug):
    
    ## create output
    clusters_union = os.path.join(pilfer_folder, "clusters.union.tsv") 
    clusters_bed = os.path.join(pilfer_folder, "merge_clusters.bed")
    clusters_merge_union = os.path.join(pilfer_folder, "merge_clusters.union.tsv")

    ## Get clusters obtained
    pilfer_code_union(pilfer_results, clusters_union)
    ## 
    bedtools_exe = set_config.get_exe("bedtools", debug)
    system_merge = "cat %s | awk \'{print $1}\' | tr \':\' \'\\t\' | tr \'-\' \'\\t\' | sort -V -k1,1 -k2,2 | grep -v \'_\' | %s merge > %s" %(clusters_union, bedtools_exe, clusters_bed)
    
    bed_merge_code = HCGB_sys.system_call(system_merge, False, True)
    if not bed_merge_code:
        print(colored("** ERROR: Something happen while calling bedtools merge for piRNA results", "red"))
        exit()
    
    ## merge clusters    
    sample_n = len(list(pilfer_results.values()))
    pilfer_code_merge(clusters_bed, clusters_union, sample_n, clusters_merge_union)
        
    ## print time stamp
    return (clusters_merge_union)

###########################################
def pilfer_code_merge(bed_file, cluster_file, sample_n, outfile):
    """
    Merges piRNA clusters from an input BED file
    
    This code is copy from PILFER software. See copyright and License details in https://github.com/rishavray/PILFER
    Original code: June 2018
    https://github.com/rishavray/PILFER/blob/master/tools/merge_cluster.py
    
    Modifications: November 2021
    - Add some comments to understand code and clarify it.
    """
    
    empty = [0 for x in range(0,sample_n)]

    clusters=[]
    for line in open(bed_file, "r"):
        line = line.strip()
        clusters.append(line.split() + empty)
    
    header=[]
    for row in open(cluster_file, "r"):
        if row.startswith("__"):
            header=row.strip().split("\t")
            header[0] = "piRNA_cluster"
            continue
    
        ## Split row 
        row = row.strip().split("\t")
            
        field = row[0].split(":")
        pos = field[1].split("-")
        pos = [int(x) for x in pos]
        for clust in clusters:
            if field[0] == clust[0] and int(clust[1]) <= pos[0] and int(clust[2]) >= pos[1]:
                for i in range(3, sample_n+3):
                    clust[i] = float(clust[i]) + float(row[i-2])
    
    outfile_hd = open(outfile, "w")
    outfile_hd.write("\t".join(header))
    outfile_hd.write("\n")
    for row in clusters:
        row[0] = row[0] + ":" + str(row[1]) + "-" + str(row[2])
        del row[1:3]
        
        outfile_hd.write("\t".join(str(elem) for elem in row))
        outfile_hd.write("\n")
    
    outfile_hd.close()

    return(True)

##########################################################3
def pilfer_code_union(list_samples_dict, outfile):
    """
    Creates union of piRNA clusters from an input BED file
    
    This code is copy from PILFER software. See copyright and License details in https://github.com/rishavray/PILFER
    Original code: June 2018
    https://github.com/rishavray/PILFER/blob/master/tools/union.py
    
    Modifications: November 2021
    - Add some comments to understand code and clarify it.
    - Add to read BED file     
    """
    
    union = {}
    i=0
    
    for idx in list_samples_dict.values():
        csvin = csv.reader(open(idx,"r"), delimiter="\t")
        for row in csvin:
            if row[0] in union:
                union[row[0]].append(row[1])
            else:
                union[row[0]] = [0 for x in range(0,i)]
                union[row[0]].append(row[1])
        for key in union:
            if len(union[key]) < i+1 :
                union[key].append(0)
        ##
        i += 1

    ## write output    
    outfile_hd = open(outfile, "w")
    csvout = csv.writer(outfile_hd, delimiter="\t")
    csvout.writerow(["__"]+ list(list_samples_dict.keys()))
    for key in union:
        csvout.writerow([key]+union[key])
    
    return(True)

##########################################################3
def pilfer_code(infile, outfile):
    """
    Predicts piRNA clusters from an input BED file
    
    This code is copy from PILFER software. See copyright and License details in https://github.com/rishavray/PILFER
    Original code: June 2018
    https://github.com/rishavray/PILFER/blob/master/tools/pilfer.py
    
    Modifications: November 2021
    - Add some comments to understand code and clarify it.
    - Add to read BED file 
    
    """
    
    ## Input example
    ##1    997178      997194    CGGGTTCATTTCCCGGATGATGCACCA::PU         1    +
    ##1    1312232    1312248    GGCGAATAGTCGGTGGCTGTAGACGCGAAACCT::PU   1    +
    ##1    1355449    1355465    ATTCGCGGATGAGCAGCGAAATGCGCAGCGTC::PU    1    +
    ##1    1909803    1909819    TTGAATTTGCCGAACCGGCTTTATGGAACC::PU      2    +
    ##1    2527004    2527020    GTGCAGATCGGCACGCCAGGGCGAAT::PU          1    -
    ##1    3473028    3473044    GCCCGACCACCAGGCCAGACCGTCAAT::PU         1    -
    ##1    3707624    3707640    AGCACGCCGCCAGCGTTCATCCTGAG::PU          5    -
    ##1    3772285    3772301    TCGGCCCACGGGCTCGTGGCGTCGGA::PU          3    +
    ##1    3772450    3772466    ATTCCGGACGAGGCGTCGCTCGATCAA::PU         1    +
    ##1    4353091    4353107    TTTCGCTTGAGGACGCTACCGATTTCATTC::PU      1    +
    
    #Variables
    sd_factor=3 ## "The factor by which the read should be away from standard deviation to be called a peak"
    count_reads = []
    chrom_dict = {}
    results = []

    infile_pt = open(infile, "r")
    #csvin = csv.reader(infile_pt, delimiter = "\t")
    
    #Reading the BED records
    for row in infile_pt:
        field=row.strip().split('\t')
        count_reads.append(float(field[4]))
        
        ## classify in a dictionary by reference sequence
        if field[0] in chrom_dict:
            chrom_dict[field[0]].append(field)
        else:
            chrom_dict[field[0]] = [field]
        
    #Mean and SD calculations
    mean = np.mean(count_reads)
    sd = np.std(count_reads)
    
    #Sorting the dictionary
    for key in chrom_dict:
        chrom_dict[key] = sorted(chrom_dict[key],key=itemgetter(1))
    #del chrom_dict['chrY']
    
    #print "Chromosome\tStart\tEnd\tScore"
    #Cluster calculation
    
    ## Fix error when I changed row -> field
    ##if (row[4] - mean)/sd >= sd_factor:
    ##    numpy.core._exceptions.UFuncTypeError: ufunc 'subtract' did not contain a loop with signature matching types (dtype('<U1'), dtype('float64')) -> None

    for key in chrom_dict:
        j = 0
        while j < len(chrom_dict[key]):
            row = chrom_dict[key][j]
            if ((int(row[4]) - mean)/sd >= sd_factor):
                start_bp = int(row[2]) - 100000
                if start_bp < 0:
                    start_bp = 0
                end_bp = int(row[2])
                score = 0
                new_score = 0
                start_read_index = -1
                end_read_index = 0
                
                cur_read_index = chrom_dict[key].index(row)
    
                #Calculate the initial score and start index
                for read in chrom_dict[key]:
                    if int(read[1]) >= int(start_bp) and int(read[1]) <= int(start_bp) + 100000:
                        score += int(read[4])
                        if start_read_index == -1:
                            start_read_index = chrom_dict[key].index(read)
                            start_bp = int(read[1])
                        end_read_index = chrom_dict[key].index(read)
    
                new_score = score
                max_start = start_read_index
                max_end = end_read_index
                #calculating optimum 100KB window
                ##for i in xrange(start_read_index+1,cur_read_index+1): ## xrange not available in python3
                for i in range(start_read_index+1,cur_read_index+1): 
                    new_score = score - int(chrom_dict[key][i-1][4])
                    while  end_read_index+1 < len(chrom_dict[key]) and int(chrom_dict[key][end_read_index +1][1]) <= int(chrom_dict[key][i][1]) + 100000 :
                        new_score += int(chrom_dict[key][end_read_index+1][4])
                        end_read_index += 1
                    
                    if new_score > score:
                        score = new_score
                        max_start = i
                        max_end = end_read_index
    
                #print (key + ":" + str(chrom_dict[key][max_start][1]) + "-" + str(chrom_dict[key][max_end][2]) + "\t" + str(score)) ## pilfer default format
                if int(chrom_dict[key][max_start][1]) < int(chrom_dict[key][max_end][2]):
                    string2print = "chr" + key + ":" + str(chrom_dict[key][max_start][1]) + "-" + str(chrom_dict[key][max_end][2]) + "\t" + str(score) + '\t+\n' ## bed format
                else:
                    string2print = "chr" + key + ":" + str(chrom_dict[key][max_end][1]) + "-" + str(chrom_dict[key][max_start][2]) + "\t" + str(score) + '\t-\n' ## bed format
                    
                results.append(string2print)
                j = max_end
            j += 1

    ## print output file
    outfile_pt = open(outfile, "w")
    for l in results:
        outfile_pt.write(l)
    outfile_pt.close()
    
    ## return information in bed format
    return (results)
            
##########################################################3
def annotate_sam(seq_id, sam_file, Debug):
    """
    This code is copy from PILFER software. See copyright and License details in https://github.com/rishavray/PILFER
    Original code: June 2018
    https://github.com/rishavray/PILFER/blob/master/tools/annotate_sam.py
    
    Modifications: November 2021
    - Add some comments to understand code and clarify it.
    - Add to read fasta file as dict and not as list as provided. 
    
    This particular scripts uses input in sam format and a file with gold piRNA sequences to identify putative and well-knonw piRNAs.
    
    """
    
    ## create output
    sam_file_out = sam_file + ".parsed"
    sam_file_write = open(sam_file_out, "w")
    
    fileReader = open(sam_file, 'r')
    
    ## Give a minimum and maximum length
    mini = 26
    maxi = 33
    
    #csvin = csv.reader(fileReader, delimiter="\t")
    #csvout = csv.writer(sam_file_write, delimiter="\t")
    for row in open(sam_file):
        ## Original
        #f = row[0].split(":")
        #row[0] = f[1]
        field=row.strip().split('\t')

        ## take into account sam header
        if (field[0].startswith('@')):
            #sam_file_write.write(row)
            continue

        seq = field[9]
        
        if (int(field[1]) & (0x10)):
            seq = HCGB_fasta.ReverseComplement(seq)
        
        ## Set putative (PU), known piRNA (PI) or none
        if seq in seq_id:
            field.append("XP:Z:PI")
            field[9] = field[9] + '::PI'
        #elif len(row[9])>=mini and len(field[9])<=maxi and int(f[0])>=100:
        elif len(field[9])>=mini and len(field[9])<=maxi:
            field.append("XP:Z:PU")
            field[9] = field[9] + '::PU'
        #else:
        ## too big or not known piRNA 
        
        ## append length in all
        field.append("XC:i:"+str(len(seq)))
    
        sam_file_write.write("\t".join(field) + "\n")
        
    ## close files
    sam_file_write.close()
    fileReader.close()

##########################################################


##########################################################
def pilfer_caller(sample_folder, name, bam_file, annot_info, threads, Debug):
    """
    Create required input files, calls pilfer to create clusters and intersect with transposon information coordinates.
    """
    
    ## create clusters using pilfer.py
    outfile = os.path.join(sample_folder, name + "-pilfer_clusters.bed")
    
    if Debug:
        HCGB_aes.debug_message("OUTFILE: " + outfile, "yellow")
        return(True)
    
    ## convert BAM to PILFER Input file
    bam_pilfer = BAMtoPILFER.process_call(bam_file, sample_folder, name, annot_info, threads, Debug)
    
    ## Call pilfer
    if bam_pilfer:
        pilfer_code(bam_pilfer, outfile)
    else:
        return False
    
    ## intersect with Transposon data
    #bedtools_caller.intersect_coordinates(outfile, annot_info["TEsmall_db"]["TE"], sample_folder, 'transposon_intersect', '', Debug)
    
    ## control if errors produced
    # return(False)
    
    return(True)


#######################################################
def pilfer_module_call(sample_folder, name, bam_file, database_folder, threads, species, Debug):
    
    ## check if previously trimmed and succeeded
    filename_stamp = sample_folder + '/.success'
    if os.path.isfile(filename_stamp):
        stamp = HCGB_time.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'pilfer'), 'yellow'))
        return(True)
    else:
        
        ## retrieved piRNA information
        annot_info = database.piRNA_info(database_folder, species, Debug)
        code_returned = pilfer_caller(sample_folder, name, bam_file, annot_info, threads, Debug)
        if code_returned:
            HCGB_time.print_time_stamp(filename_stamp)
        else:
            print ('** Sample %s failed...' %name)

#########################################
def main():
    ## this code runs when call as a single script
    parser=argparse.ArgumentParser(description='''Create piRNA analysis using PILFER''');
    
    parser.add_argument('--input', '-i', help='Input BAM file', required=True);

    parser.add_argument('--name', '-n',
                        help='Name of the sample. Default: use filename provided.', default="");
    
    parser.add_argument('--path_given', '-p',
                        help='Path to save results generated. Default: use path from filename provided.', default="");

    parser.add_argument('--database', '-db',
                        help='Absolute path for XICRA database containing piRNA files.', default="");

    parser.add_argument("-t", "--threads", type=int, help="Number of CPUs to use [Default: 2].", default=2)

    args=parser.parse_args();
    
    ## Lets convert bam2bed
    args.path_given = HCGB_files.create_folder(args.path_given)
    
    ## lets split the big file provided
    files_generated = pilfer_module_call(args.path_given, args.name, args.input, 
                                         args.database, args.threads, "hsa", Debug=True)
    
    return ()


######
if __name__== "__main__":
    main()



#usr/bin/env python
import time
import io
import os
import re
import sys
from io import open
from sys import argv
import subprocess

## ARGV
if len (sys.argv) < 2:
    print ("\nUsage:")
    print ("python3 %s gtffile folder\n" %os.path.realpath(__file__))
    exit()

gtfFile = os.path.abspath(argv[1])
folder = argv[2]

## Variables
dirname_name = os.path.dirname(gtfFile)
split_name = os.path.splitext( os.path.basename(gtfFile) )

# Open OUT
exon_gtf = folder + '/' + split_name[0] + '_exon.gtf' 
output_exon = open(exon_gtf, 'w')
miRNA_gtf = folder + '/' + split_name[0] + '_miRNA.gtf' 
output_miRNA = open(miRNA_gtf, 'w')
ncRNA_bed = folder + '/' + split_name[0] + '_ncRNA.bed' 
output_ncRNA = open(ncRNA_bed, 'w')

# Open file IN
fileHandler = open(gtfFile, "r")
while True:
    # Get next line from file
    line = fileHandler.readline().strip()
       # If line is empty then end of file reached
    if not line :
        break;
    
    if not line.startswith('#'):        
        typeID = line.split('\t')[2]
        
        if (typeID == 'exon'):
            #print (line)        
            output_exon.write(line)
            output_exon.write('\n')

        description = line.split('\t')[-1]
        description_list = line.split('\t')[-1].split(';')
        gene = []
        gene_type = []

        for items in description_list:
            gene_type_search = re.search('gene_type "(.*)"', items)
            gene_name_search = re.search('gene_name "(.*)"', items)
            if gene_type_search:
                gene_type = gene_type_search.group(1)
            elif gene_name_search:
                gene = gene_name_search.group(1)

        gene_type = str(gene_type)
        gene = str(gene)
        
        chR = line.split('\t')[0]    
        start = line.split('\t')[3]
        end = line.split('\t')[4]
        score = line.split('\t')[5]
        strand = line.split('\t')[6]
        source = line.split('\t')[1]
        extra = line.split('\t')[7]
        
        # write
        string2write = '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %(chR, start, end, gene, score, strand, source, typeID, extra, description)
        
        if gene_type in ["misc_RNA", "rRNA", "miRNA", "sRNA", "snRNA", "snoRNA", "scRNA", "macro_lncRNA", "scaRNA", "rRNA", "lincRNA"]:
            output_ncRNA.write(string2write)
            
        if (re.search('(.*)tRNA', gene_type)):
            output_ncRNA.write(string2write)        
        ## we are not writing piRNA as we would like to substract that information

        if (gene_type == 'miRNA'): ## print only miRNA gtf for miRNA analysis
            output_miRNA.write(line)
            output_miRNA.write('\n')
            
## close and finish
output_miRNA.close()
output_exon.close()
output_ncRNA.close()
fileHandler.close()            



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez                                      ##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain      ##
##########################################################
"""
Help messages for different scripts, modules
"""
from termcolor import colored
import HCGB.functions.aesthetics_functions as HCGB_aes


###############
def help_fastq_format():
    """
    Explanation of fastq format details.
    """

    HCGB_aes.boxymcboxface("Name format for samples")


    print ("Format for fastq files can be:\n")
    print ("- name.fastq.gz")
    print ("- name_1.fastq.gz, with '1' or '2' to specify the read")
    print ("- name_R2.fastq.gz, 'R1' or 'R2' to specify the read")
    print ("- name_L001_R1.fastq.gz, with the lane information as 'L00X' or '00X' after the name")
    print ("- name_L001_R1_001.fastq.gz, with '00X' at the end of the file name. This naming is used") 
    print ("- when the fastq of a sample had been cut in different files.")
    print ("- name_L001_XYZ_R1_001.fastq.gz, there can be extra info for each file, XYZ.")
    print ("\nThe input file names should be structured considering the following aspects:")
    print ("\n")

    HCGB_aes.print_sepLine("*",20,"red")
    print ("Length limitation")
    HCGB_aes.print_sepLine("*",20,"red")
    print ("There is a limitation for the sample ID ('name') of 25 characters.")
    print (colored("** XICRA provides an option to rename samples if necessary: module prep option --rename **", 'yellow'))
    print ("\n")
    
    HCGB_aes.print_sepLine("*",15,"red")
    print ("Extensions:")
    HCGB_aes.print_sepLine("*",15,"red")
    print("The suported extensions are:\n")
    print ("- name_L00x_R2.fastq\tname_L00x_R2.fq\n- name_L00x_R2.fastq.gz\tname_L00x_R2.fq.gz")
    print ("\n")

    HCGB_aes.print_sepLine("*",20,"red")
    print ("Single-end files")
    HCGB_aes.print_sepLine("*",20,"red")
    print("It is possible to provide NGS single-end files although some steps of the process could not be accomplished")
    print("using single-end files.\n")    
    print ("- name.fastq.gz")
    print ("- name.fastq")
    print ("- name.fq")
    print (colored('** Use option --single-end in the different XICRA modules. **', 'yellow'))
    print ("\n")

    HCGB_aes.print_sepLine("*",20,"red")
    print ("Paired-end files")
    HCGB_aes.print_sepLine("*",20,"red")
    print ("Paired-end files are full supported. The format for these files are:\n")
    print ("- name_1.fastq.gz, name_2.fastq.gz")
    print ("- name_R1.fastq.gz, name_R2.fastq.gz")
    print (colored('** No parameter is needed in to specify this kind of files. **', 'yellow'))
    print ("\n")    
    
    HCGB_aes.print_sepLine("*",20,"red")
    print ("Lane information")
    HCGB_aes.print_sepLine("*",20,"red")
    print("Files might contain lane information (L00x and/or 00x). XICRA")
    print("supports these names as long as follow these examples:")
    print("- name_L00x_R1.fastq.gz, name_L00x_R2.fastq.gz")
    print("- name_L00x_1.fastq.gz, name_L00x_2.fastq.gz")
    print ("\n")
    
    HCGB_aes.print_sepLine("*",20,"red")
    print ("Name extensions")
    HCGB_aes.print_sepLine("*",20,"red")
    print("It can also be the case that the reads of a sample are divided in different files.")
    print("In those cases, the files should contain a name final extension: ")
    print("- name1_L001_R1_001.fastq.gz, name1_L001_R2_001.fastq.gz")
    print("- name1_L001_R1_002.fastq.gz, name1_L001_R2_002.fastq.gz")
    print("- name1_L002_R1_001.fastq.gz, name1_L002_R2_001.fastq.gz")
    print("- name1_L002_R1_002.fastq.gz, name1_L002_R2_002.fastq.gz")
    print ("\n")
    
    HCGB_aes.print_sepLine("*",20,"red")
    print ("Extra information")
    HCGB_aes.print_sepLine("*",20,"red")
    print("In some cases, files might contain other extra information. In the following example,")
    print("XYZ is the extra information:")
    print("- name1_L001_XYZ_R1_001.fastq.gz, name1_L001_XYZ_R2_001.fastq.gz")
    print("- name1_L001_XYZ_R1_002.fastq.gz, name1_L001_XYZ_R2_002.fastq.gz")
    print ("\n")
    
    HCGB_aes.boxymcboxface("Sample identification")
        
    print ("XICRA will store the names of all the input files. After that, it will identify the samples.")
    print ("It can be the case that more than one file belong to the same sample. In order to pass this information")
    print ("to XICRA, a combination of the following parameters may be needed depending on the characteristics of the")
    print ("input file names:")
    print ("\n")

    HCGB_aes.print_sepLine("*",55,"red")
    print ("Option --include_lane:")
    HCGB_aes.print_sepLine("*",55,"red")
    print ("If you want to include lane tags (L00X, 00X) into each  each sample name (differentiate samples considering the lane):")
    print (colored('** Use option --include_lane within each module and the lane tag will also be used to identify samples. **\n', 'yellow'))
    print("However, if you want to consider as a single sample the different lanes, you need to merge the")
    print("corresponding fastq files:")
    print (colored('** Use option --merge_Reads within module prep. **\n', 'yellow'))
    
    print("As an example, considering the input files:")

    print("- name1_L001_R1.fastq.gz, name1_L001_R2.fastq.gz")
    print("- name1_L002_R1.fastq.gz, name1_L002_R2.fastq.gz")
    print("- name1_L003_R1.fastq.gz, name1_L003_R2.fastq.gz")
    print("- name1_L004_R1.fastq.gz, name1_L004_R2.fastq.gz\n")
    
    print("\t 1.By adding the option --include_lane in all modules, XICRA will identify four samples:")    
    print("\t - Sample 1: name1_L001_R1, name1_L001_R2")
    print("\t - Sample 2: name1_L002_R1, name1_L002_R2")
    print("\t - Sample 3: name1_L003_R1, name1_L003_R2")
    print("\t - Sample 4: name1_L004_R1, name1_L004_R2")
    print (colored('\t ** Remember to use option --include_lane within each module. **\n', 'yellow'))
    
    print("\t 2. By adding the options --include_lane --merge_Reads within module prep, XICRA will only")
    print("\t identify one sample, merging all the corresponding files:")
    print("\t - Sample 1: sample1_R1, sample1_R2\n")
    
    HCGB_aes.print_sepLine("*",55,"red")
    print ("Option --include_all:")
    HCGB_aes.print_sepLine("*",55,"red")
    print("In some cases, files might contain other extra information and it is necessary to use all")  
    print("file name to identify samples:")
    print (colored('** If that is the case use --include_all in al modules. **\n', 'yellow'))
    print("If you want to merge fastq files that only differ in the final extension (_001, _002, ...):")
    print (colored('** Use options --merge_Reads --include_all within module prep and only --include_all in the rest of the modules. **\n', 'yellow'))
    
    print("As an example, considering the input files:")
    print("- name1_L001_XYZ_R1_001.fastq.gz, name1_L001_XYZ_R2_001.fastq.gz")
    print("- name1_L001_XYZ_R1_002.fastq.gz, name1_L001_XYZ_R2_002.fastq.gz")
    print("- name1_L002_XYZ_R1_001.fastq.gz, name1_L002_XYZ_R2_001.fastq.gz")
    print("- name1_L002_XYZ_R1_002.fastq.gz, name1_L002_XYZ_R2_002.fastq.gz\n")
    
    print("\t 1.By adding the option --include_all in all modules, XICRA will identify four samples:")
   
    print("\t - Sample 1: name1_L001_XYZ_R1_001, name1_L001_XYZ_R2_001")
    print("\t - Sample 2: name1_L001_XYZ_R1_002, name1_L001_XYZ_R2_002")
    print("\t - Sample 3: name1_L002_XYZ_R1_001, name1_L002_XYZ_R2_001")
    print("\t - Sample 4: name1_L002_XYZ_R1_002, name1_L002_XYZ_R2_002")
    print (colored('\t ** Remember to use option --include_all within each module. **\n', 'yellow'))
    
    print("\t 1.By adding the options --include_all --merge_Reads within module prep, XICRA will identify two samples:")
    print("\t - Sample 1: name1_L001_XYZ_R1, name1_L001_XYZ_R2")
    print("\t - Sample 2: name1_L002_XYZ_R1, name1_L002_XYZ_R2")
    print (colored('\t ** Remember to use option --include_all within each module. **\n', 'yellow'))
    
def project_help():
    return ()

###############
def multiqc_help():
    return ()

###############
def print_help_adapters():
    print("Check the XICRA documentation of the module: ", "\033[4m"+ "https://xicra.readthedocs.io/en/latest/user_guide/modules/trimm.html"+"\033[0m")
    return()

###############
def help_join_reads():
    return ()

###############
def help_miRNA():
    print("Check the XICRA documentation of the module: ", "\033[4m"+ "https://xicra.readthedocs.io/en/latest/user_guide/modules/miRNA.html"+"\033[0m")
    return()

###############
def help_tRNA():
    return ()

###############
def help_piRNA():
    return ()



#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
import subprocess
import argparse
from termcolor import colored

from XICRA.config import set_config
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.system_call_functions as HCGB_sys
import HCGB.functions.time_functions as HCGB_time

#import pybedtools

#######################################################
def subtract_coordinates(file1, file2, path_given, name, options, debug):
    """
    Function to create a substraction call using bedtools
    """
        
    ## create a name
    bed_file = os.path.join(path_given, name + ".bed") 
    #bed_file_tmp = bed_file + '_tmp' 
    
    ## check if previously done
    filename_stamp = path_given + '/.' + name + '_subtract_success'
    if os.path.isfile(filename_stamp):
        if HCGB_files.is_non_zero_file(bed_file):
            stamp = HCGB_time.read_time_stamp(filename_stamp)
            print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'subtract bed annotations'), 'yellow'))
            return (bed_file)
    
    ## subtract any overlap between feature taking into account strandness
    string_options = " -s -f 0.5 -A " 
    if options:
        string_options = string_options + options
    
    ## Create call for bedtools intersect
    bedtools_exe = set_config.get_exe("bedtools", debug)
    cmd_bedtools = "%s subtract -a %s -b %s %s > %s" %(bedtools_exe, file1, file2, string_options, bed_file) 
    
    bed_code = HCGB_sys.system_call(cmd_bedtools, False, True)
    if not bed_code:
        print(colored("** ERROR: Something happen while calling bedtools subtract for job: " + name, "red"))
        exit()
        
    ## print time stamp
    HCGB_time.print_time_stamp(filename_stamp)

    return (bed_file)    

#######################################################
def intersect_coordinates(file1, file2, path_given, name, options, debug):
    """
    Function to create intersect call using bedtools
    """
        
    ## create a name
    bed_file = os.path.join(path_given, name + ".bed") 
    #bed_file_tmp = bed_file + '_tmp' 
    
    ## check if previously done
    filename_stamp = path_given + '/.' + name + '_intersect_success'
    if os.path.isfile(filename_stamp):
        if HCGB_files.is_non_zero_file(bed_file):
            stamp = HCGB_time.read_time_stamp(filename_stamp)
            print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'intersect bed annotations'), 'yellow'))
            return (bed_file)
    
    ###
    string_options = " -wa -wb "
    if options:
        string_options = string_options + options
    
    ## Create call for bedtools intersect
    bedtools_exe = set_config.get_exe("bedtools", debug)
    cmd_bedtools = "%s intersect -a %s -b %s %s > %s" %(bedtools_exe, file1, file2, string_options, bed_file) 
    
    bed_code = HCGB_sys.system_call(cmd_bedtools, False, True)

    if not bed_code:
        print(colored("** ERROR: Something happen while calling bedtools intersect for job: " + name, "red"))
        exit()
        
    ## print time stamp
    HCGB_time.print_time_stamp(filename_stamp)

    return (bed_file)    

#######################################################
def convert_bam2bed(sample, bam_file, path_given, pilfer=False, debug=False):
    """
    This functions calls bedtools to generate a conversion from BAM to BED format.
    
    It converts, sorts and collapse information retaining counts for each feature in bed format
    """

    bed_file = os.path.join(os.path.abspath(path_given), sample + ".bed") ## create a name
    bed_file_tmp = bed_file + '_tmp' 
    
    filename_stamp = path_given + '/.convert_bam2bed_success'
    if os.path.isfile(filename_stamp):
        if HCGB_files.is_non_zero_file(bed_file):
            stamp = HCGB_time.read_time_stamp(filename_stamp)
            print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, sample, 'convert_bam2bed'), 'yellow'))
            return (bed_file)
    
    ## execute conversion and count reads mapping in exact coordinates
    ## bedtools bamtobed -i bam_file > bed_file 
    ## bedtools sort -chrThenSizeA -i bed_file > bed_sort.file
    ## bedtools groupby -i bed_sort.file -o count -g 1,2,3 -c 4 > counts.bed 
    
    if pilfer:
        ## Create call in two separate calls to reduce RAM requirement
        bedtools_exe = set_config.get_exe("bedtools", debug)
        cmd_bedtools = "%s bamtobed -i %s > %s" %(bedtools_exe, bam_file, bed_file)
        bed_code = HCGB_sys.system_call(cmd_bedtools, False, True)
    
        if bed_code:
            ## print time stamp
            HCGB_time.print_time_stamp(filename_stamp)
        
            return (bed_file)
        else:
            print ("** ERROR: Some error occurred during conversion from BAM to BED... **")
            exit()
    else:
        ## Create call in two separate calls to reduce RAM requirement
        bedtools_exe = set_config.get_exe("bedtools", debug)
        cmd_bedtools = "%s bamtobed -i %s | %s groupby -o count -g 1,2,3 -c 4 > %s" %(bedtools_exe, bam_file, bedtools_exe, bed_file_tmp)
        bed_code = HCGB_sys.system_call(cmd_bedtools, False, True)
    
        if bed_code:
            cmd_bedtools2 = "%s sort -chrThenSizeA -i %s | %s groupby -o count -g 1,2,3 -c 4 > %s" %(bedtools_exe, bed_file_tmp, 
                                                                                                                    bedtools_exe, bed_file)
            bed_code2 = HCGB_sys.system_call(cmd_bedtools2, False, True)
        
        ## -----------------------------------------------
        ## Pybedtools
        ## -----------------------------------------------
        ## It might be possible to use pybedtools. We need to load bam, bed or whatever file. it is not possible to use string to absolute path
        #bed_info = pybedtools.bedtool.BedTool.bam_to_bed(bam_file)
        
        ## We might try but I guess to many RAM would be required.
        
        if not bed_code or not bed_code2:
            print ("** ERROR: Some error occurred during conversion from BAM to BED... **")
            exit()
    
    ## print time stamp
    HCGB_time.print_time_stamp(filename_stamp)

    ## remove tmp files
    os.remove(bed_file_tmp)

    return (bed_file)

## -----------------------------------------------------------
## Note on options of bedtools intersect
## -----------------------------------------------------------
# # Tool:    bedtools intersect (aka intersectBed)
# # Version: v2.29.2
# # Summary: Report overlaps between two feature files.
# # 
# # Usage:   bedtools intersect [OPTIONS] -a <bed/gff/vcf/bam> -b <bed/gff/vcf/bam>
# # 
# #     Note: -b may be followed with multiple databases and/or 
# #     wildcard (*) character(s). 
# # Options: 
# #     -wa    Write the original entry in A for each overlap.
# # 
# #     -wb    Write the original entry in B for each overlap.
# #         - Useful for knowing _what_ A overlaps. Restricted by -f and -r.
# # 
# #     -loj    Perform a "left outer join". That is, for each feature in A
# #         report each overlap with B.  If no overlaps are found, 
# #         report a NULL feature for B.
# # 
# #     -wo    Write the original A and B entries plus the number of base
# #         pairs of overlap between the two features.
# #         - Overlaps restricted by -f and -r.
# #           Only A features with overlap are reported.
# # 
# #     -wao    Write the original A and B entries plus the number of base
# #         pairs of overlap between the two features.
# #         - Overlapping features restricted by -f and -r.
# #           However, A features w/o overlap are also reported
# #           with a NULL B feature and overlap = 0.
# # 
# #     -u    Write the original A entry _once_ if _any_ overlaps found in B.
# #         - In other words, just report the fact >=1 hit was found.
# #         - Overlaps restricted by -f and -r.
# # 
# #     -c    For each entry in A, report the number of overlaps with B.
# #         - Reports 0 for A entries that have no overlap with B.
# #         - Overlaps restricted by -f, -F, -r, and -s.
# # 
# #     -C    For each entry in A, separately report the number of
# #         - overlaps with each B file on a distinct line.
# #         - Reports 0 for A entries that have no overlap with B.
# #         - Overlaps restricted by -f, -F, -r, and -s.
# # 
# #     -v    Only report those entries in A that have _no overlaps_ with B.
# #         - Similar to "grep -v" (an homage).
# # 
# #     -ubam    Write uncompressed BAM output. Default writes compressed BAM.
# # 
# #     -s    Require same strandedness.  That is, only report hits in B
# #         that overlap A on the _same_ strand.
# #         - By default, overlaps are reported without respect to strand.
# # 
# #     -S    Require different strandedness.  That is, only report hits in B
# #         that overlap A on the _opposite_ strand.
# #         - By default, overlaps are reported without respect to strand.
# # 
# #     -f    Minimum overlap required as a fraction of A.
# #         - Default is 1E-9 (i.e., 1bp).
# #         - FLOAT (e.g. 0.50)
# # 
# #     -F    Minimum overlap required as a fraction of B.
# #         - Default is 1E-9 (i.e., 1bp).
# #         - FLOAT (e.g. 0.50)
# # 
# #     -r    Require that the fraction overlap be reciprocal for A AND B.
# #         - In other words, if -f is 0.90 and -r is used, this requires
# #           that B overlap 90% of A and A _also_ overlaps 90% of B.
# # 
# #     -e    Require that the minimum fraction be satisfied for A OR B.
# #         - In other words, if -e is used with -f 0.90 and -F 0.10 this requires
# #           that either 90% of A is covered OR 10% of  B is covered.
# #           Without -e, both fractions would have to be satisfied.
# # 
# #     -split    Treat "split" BAM or BED12 entries as distinct BED intervals.
# # 
# #     -g    Provide a genome file to enforce consistent chromosome sort order
# #         across input files. Only applies when used with -sorted option.
# # 
# #     -nonamecheck    For sorted data, don't throw an error if the file has different naming conventions
# #             for the same chromosome. ex. "chr1" vs "chr01".
# # 
# #     -sorted    Use the "chromsweep" algorithm for sorted (-k1,1 -k2,2n) input.
# # 
# #     -names    When using multiple databases, provide an alias for each that
# #         will appear instead of a fileId when also printing the DB record.
# # 
# #     -filenames    When using multiple databases, show each complete filename
# #             instead of a fileId when also printing the DB record.
# # 
# #     -sortout    When using multiple databases, sort the output DB hits
# #             for each record.
# # 
# #     -bed    If using BAM input, write output as BED.
# # 
# #     -header    Print the header from the A file prior to results.
# # 
# #     -nobuf    Disable buffered output. Using this option will cause each line
# #         of output to be printed as it is generated, rather than saved
# #         in a buffer. This will make printing large output files 
# #         noticeably slower, but can be useful in conjunction with
# #         other software tools and scripts that need to process one
# #         line of bedtools output at a time.
# # 
# #     -iobuf    Specify amount of memory to use for input buffer.
# #         Takes an integer argument. Optional suffixes K/M/G supported.
# #         Note: currently has no effect with compressed files.
# # 
# # Notes: 
# #     (1) When a BAM file is used for the A file, the alignment is retained if overlaps exist,
# #     and excluded if an overlap cannot be found.  If multiple overlaps exist, they are not
# #     reported, as we are only testing for one or more overlaps.


